"""SCons.Builder

XXX

"""

#
# Copyright (c) 2001, 2002 Steven Knight
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY
# KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

__revision__ = "src/engine/SCons/Builder.py 0.D006 2002/03/28 02:47:47 software"



import os.path
import string
from Errors import UserError

import SCons.Action
import SCons.Node
import SCons.Node.FS
import SCons.Util



def Builder(**kw):
    """A factory for builder objects."""
    
    if kw.has_key('generator'):
        if kw.has_key('action'):
            raise UserError, "You must not specify both an action and a generator."
        kw['action'] = SCons.Action.CommandGenerator(kw['generator'])
        del kw['generator']

    if kw.has_key('action') and SCons.Util.is_Dict(kw['action']):
        return apply(CompositeBuilder, (), kw)
    elif kw.has_key('src_builder'):
        return apply(MultiStepBuilder, (), kw)
    else:
        return apply(BuilderBase, (), kw)



def _init_nodes(builder, env, tlist, slist):
    """Initialize lists of target and source nodes with all of
    the proper Builder information.
    """
    src_scanner = None
    if slist:
        src_key = slist[0].scanner_key()	# the file suffix
        src_scanner = env.get_scanner(src_key)
        if src_scanner:
            src_scanner = src_scanner.instance(env)

    for t in tlist:
        t.cwd = SCons.Node.FS.default_fs.getcwd()	# XXX
        t.builder_set(builder)
        t.env_set(env)
        t.add_source(slist)
        if builder.scanner:
            t.scanner_set(builder.scanner.instance(env))
        if src_scanner:
            t.src_scanner_set(src_key, src_scanner)



class BuilderBase:
    """Base class for Builders, objects that create output
    nodes (files) from input nodes (files).
    """

    def __init__(self,	name = None,
			action = None,
			prefix = '',
			suffix = '',
			src_suffix = '',
                        node_factory = SCons.Node.FS.default_fs.File,
                        target_factory = None,
                        source_factory = None,
                        scanner = None):
        if name is None:
            raise UserError, "You must specify a name for the builder."
	self.name = name
	self.action = SCons.Action.Action(action)

	self.prefix = prefix
	self.suffix = suffix
	self.src_suffix = src_suffix
        self.target_factory = target_factory or node_factory
        self.source_factory = source_factory or node_factory
        self.scanner = scanner
        if self.suffix and self.suffix[0] not in '.$':
	    self.suffix = '.' + self.suffix
        if self.src_suffix and self.src_suffix[0] not in '.$':
	    self.src_suffix = '.' + self.src_suffix

    def __cmp__(self, other):
	return cmp(self.__dict__, other.__dict__)

    def _create_nodes(self, env, target = None, source = None):
        """Create and return lists of target and source nodes.
        """
	def adjustixes(files, pre, suf):
	    ret = []
            if SCons.Util.is_String(files):
                files = string.split(files)
            if not SCons.Util.is_List(files):
	        files = [files]
	    for f in files:
                if SCons.Util.is_String(f):
		    if pre and f[:len(pre)] != pre:
                        path, fn = os.path.split(os.path.normpath(f))
                        f = os.path.join(path, pre + fn)
		    if suf:
		        if f[-len(suf):] != suf:
		            f = f + suf
		ret.append(f)
	    return ret

        tlist = SCons.Node.arg2nodes(adjustixes(target,
                                                env.subst(self.prefix),
                                                env.subst(self.suffix)),
                                     self.target_factory)

        slist = SCons.Node.arg2nodes(adjustixes(source,
                                                None,
                                                env.subst(self.src_suffix)),
                                     self.source_factory)
        return tlist, slist

    def __call__(self, env, target = None, source = None):
        tlist, slist = self._create_nodes(env, target, source)

        if len(tlist) == 1:
            _init_nodes(self, env, tlist, slist)
            tlist = tlist[0]
        else:
            _init_nodes(ListBuilder(self, env, tlist), env, tlist, slist)

        return tlist


    def execute(self, **kw):
	"""Execute a builder's action to create an output object.
	"""
	return apply(self.action.execute, (), kw)

    def get_raw_contents(self, **kw):
        """Fetch the "contents" of the builder's action.
        """
        return apply(self.action.get_raw_contents, (), kw)

    def get_contents(self, **kw):
        """Fetch the "contents" of the builder's action
        (for signature calculation).
        """
        return apply(self.action.get_contents, (), kw)

    def src_suffixes(self, env):
        if self.src_suffix != '':
            return [env.subst(self.src_suffix)]
        return []

    def targets(self, node):
        """Return the list of targets for this builder instance.

        For most normal builders, this is just the supplied node.
        """
        return [ node ]

class ListBuilder:
    """This is technically not a Builder object, but a wrapper
    around another Builder object.  This is designed to look
    like a Builder object, though, for purposes of building an
    array of targets from a single Action execution.
    """

    def __init__(self, builder, env, tlist):
        self.builder = builder
        self.scanner = builder.scanner
        self.env = env
        self.tlist = tlist

    def execute(self, **kw):
        if hasattr(self, 'status'):
            return self.status
        for t in self.tlist:
            # unlink all targets and make all directories
            # before building anything
            t.prepare()
        kw['target'] = self.tlist[0]
        self.status = apply(self.builder.execute, (), kw)
        for t in self.tlist:
            if not t is kw['target']:
                t.build()
        return self.status

    def get_raw_contents(self, **kw):
        return apply(self.builder.get_raw_contents, (), kw)

    def get_contents(self, **kw):
        return apply(self.builder.get_contents, (), kw)

    def src_suffixes(self, env):
        return self.builder.src_suffixes(env)

    def targets(self, node):
        """Return the list of targets for this builder instance.
        """
        return self.tlist

class MultiStepBuilder(BuilderBase):
    """This is a builder subclass that can build targets in
    multiple steps.  The src_builder parameter to the constructor
    accepts a builder that is called to build sources supplied to
    this builder.  The targets of that first build then become
    the sources of this builder.

    If this builder has a src_suffix supplied, then the src_builder
    builder is NOT invoked if the suffix of a source file matches
    src_suffix.
    """
    def __init__(self,  src_builder,
                        name = None,
			action = None,
			prefix = '',
			suffix = '',
			src_suffix = '',
                        node_factory = SCons.Node.FS.default_fs.File,
                        target_factory = None,
                        source_factory = None,
                        scanner=None):
        BuilderBase.__init__(self, name, action, prefix, suffix, src_suffix,
                             node_factory, target_factory, source_factory,
                             scanner)
        self.src_builder = src_builder

    def __call__(self, env, target = None, source = None):
        slist = SCons.Node.arg2nodes(source, self.source_factory)
        final_sources = []
        src_suffix = env.subst(self.src_suffix)
        sdict = {}
        for suff in self.src_builder.src_suffixes(env):
            sdict[suff] = None
        for snode in slist:
            path, ext = os.path.splitext(snode.abspath)
            if sdict.has_key(ext):
                tgt = self.src_builder(env, target = [ path ], source = snode)
                if not SCons.Util.is_List(tgt):
                    final_sources.append(tgt)
                else:
                    final_sources.extend(tgt)
            else:
                final_sources.append(snode)
        return BuilderBase.__call__(self, env, target=target,
                                    source=final_sources)

    def src_suffixes(self, env):
        return BuilderBase.src_suffixes(self, env) + \
               self.src_builder.src_suffixes(env)

class CompositeBuilder(BuilderBase):
    """This is a convenient Builder subclass that can build different
    files based on their suffixes.  For each target, this builder
    will examine the target's sources.  If they are all the same
    suffix, and that suffix is equal to one of the child builders'
    src_suffix, then that child builder will be used.  Otherwise,
    UserError is thrown."""
    def __init__(self,  name = None,
                        prefix='',
                        suffix='',
                        action = {},
                        src_builder = []):
        BuilderBase.__init__(self, name=name, prefix=prefix,
                             suffix=suffix)
        if src_builder and not SCons.Util.is_List(src_builder):
            src_builder = [src_builder]
        self.src_builder = src_builder
        self.action_dict = action
        self.sdict = {}
        self.sbuild = {}

    def __call__(self, env, target = None, source = None):
        tlist, slist = BuilderBase._create_nodes(self, env,
                                                 target=target, source=source)

        r = repr(env)
        if not self.sdict.has_key(r):
            self.sdict[r] = {}
            self.sbuild[r] = []
            for suff in self.src_suffixes(env):
                suff = env.subst(suff)
                self.sdict[r][suff] = suff
                self.sbuild[r].extend(filter(lambda x, e=env, s=suff:
                                                    e.subst(x.suffix) == s,
                                             self.src_builder))
            for sb in self.sbuild[r]:
                suff = env.subst(sb.suffix)
                for s in sb.src_suffixes(env):
                     self.sdict[r][env.subst(s)] = suff

        sufflist = map(lambda x, s=self.sdict[r]:
                              s[os.path.splitext(x.path)[1]],
                       slist)
        last_suffix = ''
        for suff in sufflist:
            if last_suffix and last_suffix != suff:
                raise UserError, "The builder for %s can only build source files of identical suffixes:  %s." % \
                      (tlist[0].path,
                       str(map(lambda t: str(t.path), tlist[0].sources)))
            last_suffix = suff

        if last_suffix:
            kw = {
                'name' : self.name,
                'action' : self.action_dict[last_suffix],
                'src_suffix' : last_suffix,
            }
            if self.sbuild[r]:
                sb = filter(lambda x, e=env, s=last_suffix:
                                   e.subst(x.suffix) == s,
                            self.sbuild[r])
                if sb:
                    kw['src_builder'] = sb[0]
            # XXX We should be able to cache this
            bld = apply(Builder, (), kw)
            for tnode in tlist:
                bld.__call__(env, target = tnode, source = slist)

        if len(tlist) == 1:
            tlist = tlist[0]
        return tlist

    def src_suffixes(self, env):
        suffixes = map(lambda k, e=env: e.subst(k), self.action_dict.keys()) + \
                   reduce(lambda x, y: x + y,
                          map(lambda b, e=env: b.src_suffixes(e),
                              self.src_builder),
                          [])
        return suffixes



"""SCons.Defaults

Builders and other things for the local site.  Here's where we'll
duplicate the functionality of autoconf until we move it into the
installation procedure or use something like qmconf.

The code that reads the registry to find MSVC components was borrowed
from distutils.msvccompiler.

"""

#
# Copyright (c) 2001, 2002 Steven Knight
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY
# KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

__revision__ = "src/engine/SCons/Defaults.py 0.D006 2002/03/28 02:47:47 software"



import os
import stat
import string
import sys

import SCons.Action
import SCons.Builder
import SCons.Errors
import SCons.Node.Alias
import SCons.Node.FS
import SCons.Scanner.C
import SCons.Scanner.Prog
import SCons.Util



CFile = SCons.Builder.Builder(name = 'CFile',
                              action = { '.l'    : '$LEXCOM',
                                         '.y'    : '$YACCCOM',
                                       },
                              suffix = '$CFILESUFFIX')

CXXFile = SCons.Builder.Builder(name = 'CXXFile',
                                action = { '.ll' : '$LEXCOM',
                                           '.yy' : '$YACCCOM',
                                         },
                                suffix = '$CXXFILESUFFIX')

CPlusPlusAction = SCons.Action.Action('$CXXCOM')

FortranAction = SCons.Action.Action('$F77COM')

FortranPPAction = SCons.Action.Action('$F77PPCOM')

Object = SCons.Builder.Builder(name = 'Object',
                               action = { '.c'   : '$CCCOM',
                                          '.C'   : CPlusPlusAction,
                                          '.cc'  : CPlusPlusAction,
                                          '.cpp' : CPlusPlusAction,
                                          '.cxx' : CPlusPlusAction,
                                          '.c++' : CPlusPlusAction,
                                          '.C++' : CPlusPlusAction,
                                          '.f'   : FortranAction,
                                          '.for' : FortranAction,
                                          '.FOR' : FortranAction,
                                          '.F'   : FortranPPAction,
                                          '.fpp' : FortranPPAction,
                                          '.FPP' : FortranPPAction,
                                        },
                               prefix = '$OBJPREFIX',
                               suffix = '$OBJSUFFIX',
                               src_builder = [CFile, CXXFile])

Program = SCons.Builder.Builder(name = 'Program',
                                action = '$LINKCOM',
                                prefix = '$PROGPREFIX',
                                suffix = '$PROGSUFFIX',
                                src_suffix = '$OBJSUFFIX',
                                src_builder = Object,
                                scanner = SCons.Scanner.Prog.ProgScan())

Library = SCons.Builder.Builder(name = 'Library',
                                action = '$ARCOM',
                                prefix = '$LIBPREFIX',
                                suffix = '$LIBSUFFIX',
                                src_suffix = '$OBJSUFFIX',
                                src_builder = Object)

LaTeXAction = SCons.Action.Action('$LATEXCOM')

DVI = SCons.Builder.Builder(name = 'DVI',
                            action = { '.tex'   : '$TEXCOM',
                                       '.ltx'   : LaTeXAction,
                                       '.latex' : LaTeXAction,
                                     },
			    # The suffix is not configurable via a
			    # construction variable like $DVISUFFIX
			    # because the output file name is
			    # hard-coded within TeX.
                            suffix = '.dvi')

PDF = SCons.Builder.Builder(name = 'PDF',
                            action = '$PDFCOM',
                            prefix = '$PDFPREFIX',
                            suffix = '$PDFSUFFIX',
                            src_suffix = '.dvi',
                            src_builder = DVI)

PostScript = SCons.Builder.Builder(name = 'PostScript',
                                   action = '$PSCOM',
                                   prefix = '$PSPREFIX',
                                   suffix = '$PSSUFFIX',
                                   src_suffix = '.dvi',
                                   src_builder = DVI)

CScan = SCons.Scanner.C.CScan()

def alias_builder(env, target, source):
    pass

Alias = SCons.Builder.Builder(name = 'Alias',
                              action = alias_builder,
                              target_factory = SCons.Node.Alias.default_ans.Alias,
                              source_factory = SCons.Node.FS.default_fs.Entry)

def get_devstudio_versions ():
    """
    Get list of devstudio versions from the Windows registry.  Return a
    list of strings containing version numbers; an exception will be raised
    if we were unable to access the registry (eg. couldn't import
    a registry-access module) or the appropriate registry keys weren't
    found.
    """

    if not SCons.Util.can_read_reg:
        raise SCons.Errors.InternalError, "No Windows registry module was found"

    K = 'Software\\Microsoft\\Devstudio'
    L = []
    for base in (SCons.Util.HKEY_CLASSES_ROOT,
                 SCons.Util.HKEY_LOCAL_MACHINE,
                 SCons.Util.HKEY_CURRENT_USER,
                 SCons.Util.HKEY_USERS):
        try:
            k = SCons.Util.RegOpenKeyEx(base,K)
            i = 0
            while 1:
                try:
                    p = SCons.Util.RegEnumKey(k,i)
                    if p[0] in '123456789' and p not in L:
                        L.append(p)
                except SCons.Util.RegError:
                    break
                i = i + 1
        except SCons.Util.RegError:
            pass

    if not L:
        raise SCons.Errors.InternalError, "DevStudio was not found."

    L.sort()
    L.reverse()
    return L

def get_msvc_path (path, version, platform='x86'):
    """
    Get a list of devstudio directories (include, lib or path).  Return
    a string delimited by ';'. An exception will be raised if unable to
    access the registry or appropriate registry keys not found.
    """

    if not SCons.Util.can_read_reg:
        raise SCons.Errors.InternalError, "No Windows registry module was found"

    if path=='lib':
        path= 'Library'
    path = string.upper(path + ' Dirs')
    K = ('Software\\Microsoft\\Devstudio\\%s\\' +
         'Build System\\Components\\Platforms\\Win32 (%s)\\Directories') % \
        (version,platform)
    for base in (SCons.Util.HKEY_CLASSES_ROOT,
                 SCons.Util.HKEY_LOCAL_MACHINE,
                 SCons.Util.HKEY_CURRENT_USER,
                 SCons.Util.HKEY_USERS):
        try:
            k = SCons.Util.RegOpenKeyEx(base,K)
            i = 0
            while 1:
                try:
                    (p,v,t) = SCons.Util.RegEnumValue(k,i)
                    if string.upper(p) == path:
                        return v
                    i = i + 1
                except SCons.Util.RegError:
                    break
        except SCons.Util.RegError:
            pass

    # if we got here, then we didn't find the registry entries:
    raise SCons.Errors.InternalError, "%s was not found in the registry."%path

def get_msdev_dir(version):
    """Returns the root directory of the MSDev installation from the
    registry if it can be found, otherwise we guess."""
    if SCons.Util.can_read_reg:
        K = ('Software\\Microsoft\\Devstudio\\%s\\' +
             'Products\\Microsoft Visual C++') % \
             version
        for base in (SCons.Util.HKEY_LOCAL_MACHINE,
                     SCons.Util.HKEY_CURRENT_USER):
            try:
                k = SCons.Util.RegOpenKeyEx(base,K)
                val, tok = SCons.Util.RegQueryValueEx(k, 'ProductDir')
                return os.path.split(val)[0]
            except SCons.Util.RegError:
                pass

def make_win32_env_from_paths(include, lib, path):
    """
    Build a dictionary of construction variables for a win32 platform.
    include - include path
    lib - library path
    path - executable path
    """
    return {
        'CC'         : 'cl',
        'CCFLAGS'    : '/nologo',
        'CCCOM'      : '$CC $CCFLAGS $CPPFLAGS $_INCFLAGS /c $SOURCES /Fo$TARGET',
        'CFILESUFFIX' : '.c',
        'CXX'        : '$CC',
        'CXXFLAGS'   : '$CCFLAGS',
        'CXXCOM'     : '$CXX $CXXFLAGS $CPPFLAGS $_INCFLAGS /c $SOURCES /Fo$TARGET',
        'CXXFILESUFFIX' : '.cc',
        'F77'        : 'g77',
        'F77FLAGS'   : '',
        'F77COM'     : '$F77 $F77FLAGS $_INCFLAGS -c -o $TARGET $SOURCES',
        'F77PPCOM'   : '$F77 $F77FLAGS $CPPFLAGS $_INCFLAGS -c -o $TARGET $SOURCES',
        'LINK'       : 'link',
        'LINKFLAGS'  : '/nologo',
        'LINKCOM'    : '$LINK $LINKFLAGS /OUT:$TARGET $_LIBDIRFLAGS $_LIBFLAGS $SOURCES',
        'AR'         : 'lib',
        'ARFLAGS'    : '/nologo',
        'ARCOM'      : '$AR $ARFLAGS /OUT:$TARGET $SOURCES',
        'LEX'        : 'lex',
        'LEXFLAGS'   : '',
        'LEXCOM'     : '$LEX $LEXFLAGS -t $SOURCES > $TARGET',
        'YACC'       : 'yacc',
        'YACCFLAGS'  : '',
        'YACCCOM'    : '$YACC $YACCFLAGS -o $TARGET $SOURCES',
        'TEX'        : 'tex',
        'TEXFLAGS'   : '',
        'TEXCOM'     : '$TEX $TEXFLAGS $SOURCES',
        'LATEX'      : 'latex',
        'LATEXFLAGS' : '',
        'LATEXCOM'   : '$LATEX $LATEXFLAGS $SOURCES',
        'DVIPDF'     : 'dvipdf',
        'DVIPDFFLAGS' : '',
        'PDFCOM'     : '$DVIPDF $DVIPDFFLAGS $SOURCES $TARGET',
        'PDFPREFIX'  : '',
        'PDFSUFFIX'  : '.pdf',
        'DVIPS'      : 'dvips',
        'DVIPSFLAGS' : '',
        'PSCOM'      : '$DVIPS $DVIPSFLAGS -o $TARGET $SOURCES',
        'PSPREFIX'   : '',
        'PSSUFFIX'   : '.ps',
        'BUILDERS'   : [Alias, CFile, CXXFile, DVI, Library, Object,
                        PDF, PostScript, Program],
        'SCANNERS'   : [CScan],
        'OBJPREFIX'  : '',
        'OBJSUFFIX'  : '.obj',
        'PROGPREFIX' : '',
        'PROGSUFFIX' : '.exe',
        'LIBPREFIX'  : '',
        'LIBSUFFIX'  : '.lib',
        'LIBDIRPREFIX'          : '/LIBPATH:',
        'LIBDIRSUFFIX'          : '',
        'LIBLINKPREFIX'         : '',
        'LIBLINKSUFFIX'         : '$LIBSUFFIX',
        'INCPREFIX'             : '/I',
        'INCSUFFIX'             : '',
        'ENV'        : {
            'INCLUDE'  : include,
            'LIB'      : lib,
            'PATH'     : path,
                'PATHEXT' : '.COM;.EXE;.BAT;.CMD',
            },
        }

def make_win32_env(version):
    """
    Build a dictionary of construction variables for a win32 platform.
    ver - the version string of DevStudio to use (e.g. "6.0")
    """
    return make_win32_env_from_paths(get_msvc_path("include", version),
                                     get_msvc_path("lib", version),
                                     get_msvc_path("path", version)
                                     + ";" + os.environ['PATH'])


if os.name == 'posix':

    arcom = '$AR $ARFLAGS $TARGET $SOURCES'
    ranlib = 'ranlib'
    if SCons.Util.WhereIs(ranlib):
        arcom = arcom + '\n$RANLIB $RANLIBFLAGS $TARGET'

    ConstructionEnvironment = {
        'CC'         : 'cc',
        'CCFLAGS'    : '',
        'CCCOM'      : '$CC $CCFLAGS $CPPFLAGS $_INCFLAGS -c -o $TARGET $SOURCES',
        'CFILESUFFIX' : '.c',
        'CXX'        : 'c++',
        'CXXFLAGS'   : '$CCFLAGS',
        'CXXCOM'     : '$CXX $CXXFLAGS $CPPFLAGS $_INCFLAGS -c -o $TARGET $SOURCES',
        'CXXFILESUFFIX' : '.cc',
        'F77'        : 'g77',
        'F77FLAGS'   : '',
        'F77COM'     : '$F77 $F77FLAGS $_INCFLAGS -c -o $TARGET $SOURCES',
        'F77PPCOM'   : '$F77 $F77FLAGS $CPPFLAGS $_INCFLAGS -c -o $TARGET $SOURCES',
        'LINK'       : '$CXX',
        'LINKFLAGS'  : '',
        'LINKCOM'    : '$LINK $LINKFLAGS -o $TARGET $SOURCES $_LIBDIRFLAGS $_LIBFLAGS',
        'AR'         : 'ar',
        'ARFLAGS'    : 'r',
        'RANLIB'     : ranlib,
        'RANLIBFLAGS' : '',
        'ARCOM'      : arcom,
        'LEX'        : 'lex',
        'LEXFLAGS'   : '',
        'LEXCOM'     : '$LEX $LEXFLAGS -t $SOURCES > $TARGET',
        'YACC'       : 'yacc',
        'YACCFLAGS'  : '',
        'YACCCOM'    : '$YACC $YACCFLAGS -o $TARGET $SOURCES',
        'TEX'        : 'tex',
        'TEXFLAGS'   : '',
        'TEXCOM'     : '$TEX $TEXFLAGS $SOURCES',
        'LATEX'      : 'latex',
        'LATEXFLAGS' : '',
        'LATEXCOM'   : '$LATEX $LATEXFLAGS $SOURCES',
        'DVIPDF'     : 'dvipdf',
        'PDFCOM'     : '$DVIPDF $DVIPDFFLAGS $SOURCES $TARGET',
        'PDFPREFIX'  : '',
        'PDFSUFFIX'  : '.pdf',
        'DVIPS'      : 'dvips',
        'PSCOM'      : '$DVIPS $DVIPSFLAGS -o $TARGET $SOURCES',
        'PSPREFIX'   : '',
        'PSSUFFIX'   : '.ps',
        'BUILDERS'   : [Alias, CFile, CXXFile, DVI, Library, Object,
                        PDF, PostScript, Program],
        'SCANNERS'   : [CScan],
        'OBJPREFIX'  : '',
        'OBJSUFFIX'  : '.o',
        'PROGPREFIX' : '',
        'PROGSUFFIX' : (sys.platform == 'cygwin') and '.exe' or '',
        'LIBPREFIX'  : 'lib',
        'LIBSUFFIX'  : '.a',
        'LIBDIRPREFIX'          : '-L',
        'LIBDIRSUFFIX'          : '',
        'LIBLINKPREFIX'         : '-l',
        'LIBLINKSUFFIX'         : '',
        'INCPREFIX'             : '-I',
        'INCSUFFIX'             : '',
        'ENV'        : { 'PATH' : '/usr/local/bin:/bin:/usr/bin' },
    }

elif os.name == 'nt':
    versions = None
    try:
        versions = get_devstudio_versions()
        ConstructionEnvironment = make_win32_env(versions[0]) #use highest version
    except (SCons.Util.RegError, SCons.Errors.InternalError):
        # Could not get the configured directories from the registry.
        # However, the configured directories only appear if the user
        # changes them from the default.  Therefore, we'll see if
        # we can get the path to the MSDev base installation from
        # the registry and deduce the default directories.
        MVSdir = None
        if versions:
            MVSdir = get_msdev_dir(versions[0])
        if MVSdir:
            MVSVCdir = r'%s\VC98' % MVSdir
            MVSCommondir = r'%s\Common' % MVSdir
            try:
                extra_path = os.pathsep + os.environ['PATH']
            except KeyError:
                extra_path = ''
            ConstructionEnvironment = make_win32_env_from_paths(
                r'%s\atl\include;%s\mfc\include;%s\include' % (MVSVCdir, MVSVCdir, MVSVCdir),
                r'%s\mfc\lib;%s\lib' % (MVSVCdir, MVSVCdir),
                (r'%s\MSDev98\Bin;%s\Bin' % (MVSCommondir, MVSVCdir)) + extra_path)
        else:
            # The DevStudio environment variables don't exist,
            # so just use the variables from the source environment.
            MVSdir = r'C:\Program Files\Microsoft Visual Studio'
            MVSVCdir = r'%s\VC98' % MVSdir
            MVSCommondir = r'%s\Common' % MVSdir
            try:
                include_path = os.environ['INCLUDE']
            except KeyError:
                include_path = ''
            try:
                lib_path = os.environ['LIB']
            except KeyError:
                lib_path = ''
            try:
                exe_path = os.environ['PATH']
            except KeyError:
                exe_path = ''
            ConstructionEnvironment = make_win32_env_from_paths(
                include_path,
                lib_path,
                exe_path)




"""SCons.Node

The Node package for the SCons software construction utility.

"""

#
# Copyright (c) 2001, 2002 Steven Knight
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY
# KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

__revision__ = "src/engine/SCons/Node/__init__.py 0.D006 2002/03/28 02:47:47 software"



import string
import types
import copy
import sys

from SCons.Errors import BuildError
import SCons.Util

# Node states
#
# These are in "priority" order, so that the maximum value for any
# child/dependency of a node represents the state of that node if
# it has no builder of its own.  The canonical example is a file
# system directory, which is only up to date if all of its children
# were up to date.
pending = 1
executing = 2
up_to_date = 3
executed = 4
failed = 5

class Node:
    """The base Node class, for entities that we know how to
    build, or use to build other Nodes.
    """

    def __init__(self):
        self.sources = []       # source files used to build node
        self.depends = []       # explicit dependencies (from Depends)
        self.implicit = {}	# implicit (scanned) dependencies
        self.ignore = []	# dependencies to ignore
        self.parents = {}
        self.wkids = None       # Kids yet to walk, when it's an array
        self.builder = None
        self.scanner = None     # explicit scanner from this node's Builder
        self.scanned = {}       # cached scanned values
        self.src_scanners = {}  # scanners for this node's source files
        self.env = None
        self.state = None
        self.bsig = None
        self.csig = None
        self.use_signature = 1
        self.precious = None
        self.found_includes = {}

    def build(self):
        """Actually build the node.   Return the status from the build."""
	if not self.builder:
	    return None
        try:
            # If this Builder instance has already been called,
            # there will already be an associated status.
            stat = self.builder.status
        except AttributeError:
            try:
                dict = copy.copy(self.env.Dictionary())
                if hasattr(self, 'dir'):
                    auto = self.env.autogenerate(dir = self.dir)
                else:
                    auto = self.env.autogenerate()
                dict.update(auto)
                stat = self.builder.execute(env = dict,
                                            target = self,
                                            source = self.sources)
            except:
                raise BuildError(self, "Exception",
                                 sys.exc_type,
                                 sys.exc_value,
                                 sys.exc_traceback)
        if stat:
            raise BuildError(node = self, errstr = "Error %d" % stat)

        self.found_includes = {}

        # If we successfully build a node, then we need to rescan for
        # implicit dependencies, since it might have changed on us.
        self.scanned = {}

        return stat

    def builder_set(self, builder):
	self.builder = builder

    def builder_sig_adapter(self):
        """Create an adapter for calculating a builder's signature.

        The underlying signature class will call get_contents()
        to fetch the signature of a builder, but the actual
        content of that signature depends on the node and the
        environment (for construction variable substitution),
        so this adapter provides the right glue between the two.
        """
        class Adapter:
            def __init__(self, node):
                self.node = node
            def get_contents(self):
                dict = self.node.env.Dictionary()
                dict.update(self.node.env.autogenerate())
                try:
                    dir = self.node.getcwd()
                except AttributeError:
                    dir = None
                return self.node.builder.get_contents(env = dict)
        return Adapter(self)

    def scanner_set(self, scanner):
        self.scanner = scanner

    def src_scanner_set(self, key, scanner):
        self.src_scanners[key] = scanner

    def src_scanner_get(self, key):
        return self.src_scanners.get(key, None)

    def scan(self, scanner = None):
        if not scanner:
            scanner = self.scanner
        self.scanned[scanner] = 1

    def scanner_key(self):
        return None

    def env_set(self, env, safe=0):
        if safe and self.env:
            return
	self.env = env

    def get_bsig(self):
        """Get the node's build signature (based on the signatures
        of its dependency files and build information)."""
        return self.bsig

    def set_bsig(self, bsig):
        """Set the node's build signature (based on the signatures
        of its dependency files and build information)."""
        self.bsig = bsig

    def get_csig(self):
        """Get the signature of the node's content."""
        return self.csig

    def set_csig(self, csig):
        """Set the signature of the node's content."""
        self.csig = csig

    def store_sigs(self):
        """Make the signatures permanent (that is, store them in the
        .sconsign file or equivalent)."""
        pass

    def set_precious(self, precious = 1):
        """Set the Node's precious value."""
        self.precious = precious

    def prepare(self):
        """Prepare for this Node to be created:  no-op by default."""
        pass

    def add_dependency(self, depend):
	"""Adds dependencies. The depend argument must be a list."""
        self._add_child(self.depends, depend)

    def add_ignore(self, depend):
        """Adds dependencies to ignore. The depend argument must be a list."""
        self._add_child(self.ignore, depend)

    def add_source(self, source):
	"""Adds sources. The source argument must be a list."""
        self._add_child(self.sources, source)

    def add_implicit(self, implicit, key):
        """Adds implicit (scanned) dependencies. The implicit
        argument must be a list."""
        if not self.implicit.has_key(key):
             self.implicit[key] = []
        self._add_child(self.implicit[key], implicit)

    def _add_child(self, collection, child):
        """Adds 'child' to 'collection'. The 'child' argument must be a list"""
        if type(child) is not type([]):
            raise TypeError("child must be a list")
	child = filter(lambda x, s=collection: x not in s, child)
	if child:
	    collection.extend(child)

        for c in child:
            c.parents[self] = 1

    def add_wkid(self, wkid):
        """Add a node to the list of kids waiting to be evaluated"""
        if self.wkids != None:
            self.wkids.append(wkid)

    def children(self, scanner):
        """Return a list of the node's direct children, minus those
        that are ignored by this node."""
        return filter(lambda x, i=self.ignore: x not in i,
                      self.all_children(scanner))

    def all_children(self, scanner):
        """Return a list of all the node's direct children."""
        #XXX Need to remove duplicates from this
        if not self.implicit.has_key(scanner):
            self.scan(scanner)
        if scanner:
            implicit = self.implicit[scanner]
        else:
            implicit = reduce(lambda x, y: x + y, self.implicit.values(), [])
        return self.sources + self.depends + implicit

    def get_parents(self):
        return self.parents.keys()

    def set_state(self, state):
        self.state = state

    def get_state(self):
        return self.state

    def current(self):
        return None

    def children_are_executed(self, scanner):
        return reduce(lambda x,y: ((y.get_state() == executed
                                   or y.get_state() == up_to_date)
                                   and x),
                      self.children(scanner),
                      1)

def get_children(node, parent): return node.children(None)
def ignore_cycle(node, stack): pass
def do_nothing(node, parent): pass

class Walker:
    """An iterator for walking a Node tree.

    This is depth-first, children are visited before the parent.
    The Walker object can be initialized with any node, and
    returns the next node on the descent with each next() call.
    'kids_func' is an optional function that will be called to
    get the children of a node instead of calling 'children'. 
    'cycle_func' is an optional function that will be called
    when a cycle is detected.
    
    This class does not get caught in node cycles caused, for example,
    by C header file include loops.
    """
    def __init__(self, node, kids_func=get_children,
                             cycle_func=ignore_cycle,
                             eval_func=do_nothing):
        self.kids_func = kids_func
        self.cycle_func = cycle_func
        self.eval_func = eval_func
        node.wkids = copy.copy(kids_func(node, None))
        self.stack = [node]
        self.history = {} # used to efficiently detect and avoid cycles
        self.history[node] = None

    def next(self):
	"""Return the next node for this walk of the tree.

	This function is intentionally iterative, not recursive,
	to sidestep any issues of stack size limitations.
	"""

	while self.stack:
            if self.stack[-1].wkids:
                node = self.stack[-1].wkids.pop(0)
                if not self.stack[-1].wkids:
                    self.stack[-1].wkids = None
                if self.history.has_key(node):
                    self.cycle_func(node, self.stack)
                else:
                    node.wkids = copy.copy(self.kids_func(node, self.stack[-1]))
                    self.stack.append(node)
                    self.history[node] = None
            else:
                node = self.stack.pop()
                del self.history[node]
                if node:
                    if self.stack:
                        parent = self.stack[-1]
                    else:
                        parent = None
                    self.eval_func(node, parent)
                return node

    def is_done(self):
        return not self.stack


arg2nodes_lookups = []


def arg2nodes(arg, node_factory=None):
    """This function converts a string or list into a list of Node instances.
    It follows the rules outlined in the SCons design document by accepting
    any of the following inputs:
        - A single string containing names separated by spaces. These will be
          split apart at the spaces.
        - A single Node instance,
        - A list containingg either strings or Node instances. Any strings
          in the list are not split at spaces.
    In all cases, the function returns a list of Node instances."""

    narg = arg
    if SCons.Util.is_String(arg):
        narg = string.split(arg)
    elif not SCons.Util.is_List(arg):
        narg = [arg]

    nodes = []
    for v in narg:
        if SCons.Util.is_String(v):
            n = None
            for l in arg2nodes_lookups:
                n = l(v)
                if not n is None:
                    break
            if not n is None:
                nodes.append(n)
            elif node_factory:
                nodes.append(node_factory(v))
        # Do we enforce the following restriction?  Maybe, but it
        # would also restrict what we can do to allow people to
        # use the engine with alternate Node implementations...
        #elif not issubclass(v.__class__, Node):
        #    raise TypeError
        else:
            nodes.append(v)

    return nodes



"""SCons.Scanner.C

This module implements the depenency scanner for C/C++ code. 

"""

#
# Copyright (c) 2001, 2002 Steven Knight
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY
# KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

__revision__ = "src/engine/SCons/Scanner/C.py 0.D006 2002/03/28 02:47:47 software"


import copy
import os.path
import re

import SCons.Node
import SCons.Node.FS
import SCons.Scanner
import SCons.Util

include_re = re.compile('^[ \t]*#[ \t]*include[ \t]+(<|")([\\w./\\\\]+)(>|")', re.M)

include_cache = {}

def CScan(fs = SCons.Node.FS.default_fs):
    """Return a prototype Scanner instance for scanning source files
    that use the C pre-processor"""
    cs = CScanner(scan, "CScan", [fs, ()],
                  [".c", ".C", ".cxx", ".cpp", ".c++", ".cc",
                   ".h", ".H", ".hxx", ".hpp", ".hh",
                   ".F", ".fpp", ".FPP"])
    cs.fs = fs
    return cs

class CScanner(SCons.Scanner.Recursive):
    def __init__(self, *args, **kw):
        apply(SCons.Scanner.Recursive.__init__, (self,) + args, kw)
        self.hash = None
        self.pathscanners = {}

    def instance(self, env):
        """
        Return a unique instance of a C scanner object for a
        given environment.
        """
        try:
            dirs = tuple(SCons.Node.arg2nodes(env.Dictionary('CPPPATH'),
                                              self.fs.Dir))
        except:
            dirs = ()
        if not self.pathscanners.has_key(dirs):
            clone = copy.copy(self)
            clone.hash = dirs
            clone.argument = [self.fs, dirs]	# XXX reaching into object
            self.pathscanners[dirs] = clone
        return self.pathscanners[dirs]

    def __hash__(self):
        return hash(self.hash)

def scan(node, env, args = [SCons.Node.FS.default_fs, ()]):
    """
    scan(node, Environment) -> [node]

    the C/C++ dependency scanner function

    This function is intentionally simple. There are two rules it
    follows:
    
    1) #include <foo.h> - search for foo.h in CPPPATH followed by the
        directory 'filename' is in
    2) #include \"foo.h\" - search for foo.h in the directory 'filename' is
       in followed by CPPPATH

    These rules approximate the behaviour of most C/C++ compilers.

    This scanner also ignores #ifdef and other preprocessor conditionals, so
    it may find more depencies than there really are, but it never misses
    dependencies.
    """

    fs, cpppath = args
    nodes = []

    try:
        nodes = node.found_includes[cpppath]
    except KeyError:
        if node.exists():

            # cache the includes list in node so we only scan it once:
            if hasattr(node, 'includes'):
                includes = node.includes
            else:
                includes = include_re.findall(node.get_contents())
                node.includes = includes

            source_dir = node.get_dir()
            
            for include in includes:
                if include[0] == '"':
                    n = SCons.Node.FS.find_file(include[1],
                                                (source_dir,) + cpppath,
                                                fs.File)
                else:
                    n = SCons.Node.FS.find_file(include[1],
                                                cpppath + (source_dir,),
                                                fs.File)

                if not n is None:
                    nodes.append(n)
        node.found_includes[cpppath] = nodes

    # Schwartzian transform from the Python FAQ Wizard
    def st(List, Metric):
        def pairing(element, M = Metric):
            return (M(element), element)
        def stripit(pair):
            return pair[1]
        paired = map(pairing, List)
        paired.sort()
        return map(stripit, paired)

    def normalize(node):
        return str(node)

    return st(nodes, normalize)





"""SCons.Sig.MD5

The MD5 signature package for the SCons software construction
utility.

"""

#
# Copyright (c) 2001, 2002 Steven Knight
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY
# KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

__revision__ = "src/engine/SCons/Sig/MD5.py 0.D006 2002/03/28 02:47:47 software"

import imp
import string

# Force Python to load the builtin "md5" module.  If we do this with a
# normal import statement, then case-insensitive systems (Win32) get
# confused and thinks there's a case mismatch with *this* MD5.py module.
file, name, desc = imp.find_module('md5')
try:
    md5 = imp.load_module('md5', file, name, desc)
finally:
    if file:
        file.close()

def current(new, old):
    """Return whether a new signature is up-to-date with
    respect to an old signature.
    """
    return new == old

def hexdigest(s):
    """Return a signature as a string of hex characters.
    """
    # NOTE:  This routine is a method in the Python 2.0 interface
    # of the native md5 module, but we want SCons to operate all
    # the way back to at least Python 1.5.2, which doesn't have it.
    h = string.hexdigits
    r = ''
    for c in s:
	i = ord(c)
	r = r + h[(i >> 4) & 0xF] + h[i & 0xF]
    return r

def collect(signatures):
    """
    Collect a list of signatures into an aggregate signature.

    signatures - a list of signatures
    returns - the aggregate signature
    """
    if len(signatures) == 1:
	return signatures[0]
    else:
        contents = string.join(signatures, ', ')
	return hexdigest(md5.new(contents).digest())

def signature(obj):
    """Generate a signature for an object
    """
    try:
        contents = str(obj.get_contents())
    except AttributeError, e:
        raise AttributeError, \
	      "unable to fetch contents of '%s': %s" % (str(obj), e)
    return hexdigest(md5.new(contents).digest())

def to_string(signature):
    """Convert a signature to a string"""
    return signature

def from_string(string):
    """Convert a string to a signature"""
    return string



"""SCons.Scanner

The Scanner package for the SCons software construction utility.

"""

#
# Copyright (c) 2001, 2002 Steven Knight
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY
# KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

__revision__ = "src/engine/SCons/Scanner/__init__.py 0.D006 2002/03/28 02:47:47 software"

__version__ = "0.06"


import SCons.Node.FS
import SCons.Util


class _Null:
    pass

# This is used instead of None as a default argument value so None can be
# used as an actual argument value.
_null = _Null

class Base:
    """
    The base class for dependency scanners.  This implements
    straightforward, single-pass scanning of a single file.
    """

    def __init__(self,
                 function,
                 name = "NONE",
                 argument = _null,
                 skeys = [],
                 node_factory = SCons.Node.FS.default_fs.File):
        """
        Construct a new scanner object given a scanner function.

        'function' - a scanner function taking two or three arguments and
        returning a list of strings.

        'name' - a name for identifying this scanner object.

        'argument' - an optional argument that will be passed to the
        scanner function if it is given.

        'skeys; - an optional list argument that can be used to determine
        which scanner should be used for a given Node. In the case of File
        nodes, for example, the 'skeys' would be file suffixes.

        The scanner function's first argument will be the name of a file
        that should be scanned for dependencies, the second argument will
        be an Environment object, the third argument will be the value
        passed into 'argument', and the returned list should contain the
        Nodes for all the direct dependencies of the file.

        Examples:

        s = Scanner(my_scanner_function)
        
        s = Scanner(function = my_scanner_function)

        s = Scanner(function = my_scanner_function, argument = 'foo')

        """

        # Note: this class could easily work with scanner functions that take
        # something other than a filename as an argument (e.g. a database
        # node) and a dependencies list that aren't file names. All that
        # would need to be changed is the documentation.

        self.function = function
        self.name = name
        self.argument = argument
        self.skeys = skeys
        self.node_factory = node_factory

    def scan(self, node, env):
        """
        This method scans a single object. 'node' is the node
        that will be passed to the scanner function, and 'env' is the
        environment that will be passed to the scanner function. A list of
        direct dependency nodes for the specified node will be returned.
        """

        if not self.argument is _null:
            list = self.function(node, env, self.argument)
        else:
            list = self.function(node, env)
        kw = {}
        if hasattr(node, 'dir'):
            kw['directory'] = node.dir
        nodes = []
        for l in list:
            if not isinstance(l, SCons.Node.FS.Entry):
                l = apply(self.node_factory, (l,), kw)
            nodes.append(l)
        return nodes

    def instance(self, env):
        """
        Return an instance of a Scanner object for use in scanning.

        In the base class, we just return the scanner itself.
        Other Scanner classes may use this to clone copies and/or
        return unique instances as needed.
        """
        return self

    def __cmp__(self, other):
        return cmp(self.__dict__, other.__dict__)

    def __hash__(self):
        return hash(None)

class Recursive(Base):
    """
    The class for recursive dependency scanning.  This will
    re-scan any new files returned by each call to the
    underlying scanning function, and return the aggregate
    list of all dependencies.
    """

    def scan(self, node, env):
        """
        This method does the actual scanning. 'node' is the node
        that will be passed to the scanner function, and 'env' is the
        environment that will be passed to the scanner function. An
        aggregate list of dependency nodes for the specified filename
        and any of its scanned dependencies will be returned.
        """

        nodes = [node]
        seen = {node : 0}
        deps = []
        while nodes:
            n = nodes.pop(0)
            d = filter(lambda x, seen=seen: not seen.has_key(x),
                       Base.scan(self, n, env))
            if d:
                deps.extend(d)
                nodes.extend(d)
                for n in d:
                    seen[n] = 0
        return deps



"""SCons.Sig

The Signature package for the scons software construction utility.

"""

#
# Copyright (c) 2001, 2002 Steven Knight
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY
# KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

__revision__ = "src/engine/SCons/Sig/__init__.py 0.D006 2002/03/28 02:47:47 software"

import os
import os.path
import string
import SCons.Node

#XXX Get rid of the global array so this becomes re-entrant.
sig_files = []

def write():
    global sig_files
    for sig_file in sig_files:
        sig_file.write()

class SConsignFile:
    """
    Encapsulates reading and writing a .sconsign file.
    """

    def __init__(self, dir, module):
        """
        dir - the directory for the file
        module - the signature module being used
        """
        
        self.dir = dir
        self.module = module
        self.sconsign = os.path.join(dir.path, '.sconsign')
        self.entries = {}
        self.dirty = None
                    
        try:
            file = open(self.sconsign, 'rt')
        except:
            pass
        else:
            for line in file.readlines():
                filename, rest = map(string.strip, string.split(line, ":"))
                self.entries[filename] = rest

        global sig_files
        sig_files.append(self)

    def get(self, filename):
        """
        Get the .sconsign entry for a file

        filename - the filename whose signature will be returned
        returns - (timestamp, bsig, csig)
        """
        
        try:
            arr = map(string.strip, string.split(self.entries[filename], " "))
        except KeyError:
            return (None, None, None)
        try:
            if arr[1] == '-': bsig = None
            else:             bsig = self.module.from_string(arr[1])
        except IndexError:
            bsig = None
        try:
            if arr[2] == '-': csig = None
            else:             csig = self.module.from_string(arr[2])
        except IndexError:
            csig = None
        return (int(arr[0]), bsig, csig)

    def set(self, filename, timestamp, bsig = None, csig = None):
        """
        Set the .sconsign entry for a file

        filename - the filename whose signature will be set
        timestamp - the timestamp
        module - the signature module being used
        bsig - the file's build signature
        csig - the file's content signature
        """
        if bsig is None: bsig = '-'
        else:            bsig = self.module.to_string(bsig)
        if csig is None: csig = ''
        else:            csig = ' ' + self.module.to_string(csig)
        self.entries[filename] = "%d %s%s" % (timestamp, bsig, csig)
        self.dirty = 1

    def write(self):
        """
        Write the .sconsign file to disk.

        Try to write to a temporary file first, and rename it if we
        succeed.  If we can't write to the temporary file, it's
        probably because the directory isn't writable (and if so,
        how did we build anything in this directory, anyway?), so
        try to write directly to the .sconsign file as a backup.
        If we can't rename, try to copy the temporary contents back
        to the .sconsign file.  Either way, always try to remove
        the temporary file at the end.
        """
        if self.dirty:
            temp = os.path.join(self.dir.path, '.scons%d' % os.getpid())
            try:
                file = open(temp, 'wt')
                fname = temp
            except:
                file = open(self.sconsign, 'wt')
                fname = self.sconsign
            keys = self.entries.keys()
            keys.sort()
            for name in keys:
                file.write("%s: %s\n" % (name, self.entries[name]))
            file.close()
            if fname != self.sconsign:
                try:
                    mode = os.stat(self.sconsign)[0]
                    os.chmod(self.sconsign, 0666)
                    os.unlink(self.sconsign)
                except:
                    pass
                try:
                    os.rename(fname, self.sconsign)
                except:
                    open(self.sconsign, 'wb').write(open(fname, 'rb').read())
                    os.chmod(self.sconsign, mode)
            try:
                os.unlink(temp)
            except:
                pass


class Calculator:
    """
    Encapsulates signature calculations and .sconsign file generating
    for the build engine.
    """

    def __init__(self, module):
        """
        Initialize the calculator.

        module - the signature module to use for signature calculations
        """
        self.module = module

    def bsig(self, node):
        """
        Generate a node's build signature, the digested signatures
        of its dependency files and build information.

        node - the node whose sources will be collected
        returns - the build signature

        This no longer handles the recursive descent of the
        node's children's signatures.  We expect that they're
        already built and updated by someone else, if that's
        what's wanted.
        """
        if not node.use_signature:
            return None
        #XXX If configured, use the content signatures from the
        #XXX .sconsign file if the timestamps match.

        bsig = node.get_bsig()
        if not bsig is None:
            return bsig

        # Collect the signatures for ALL the nodes that this
        # node depends on. Just collecting the direct
        # dependants is not good enough, because
        # the signature of a non-derived file does
        # not include the signatures of its psuedo-sources
        # (e.g. the signature for a .c file does not include
        # the signatures of the .h files that it includes).

        # However, we do NOT want to walk dependencies of non-
        # derived files, because calling get_signature() on the
        # derived nodes will in turn call bsig() again and do that
        # for us.  Hence:
        sigs = []
        def non_derived(n, parent, myself=node):
            if not n.builder or n is myself:
                return filter(lambda x, i=myself.ignore: x not in i,
                              n.all_children(None))
            return []
        def get_sig(n, parent, self=self, myself=node, sigs=sigs):
            if not n is myself:
                sigs.append(self.get_signature(n))
        walker = SCons.Node.Walker(node, non_derived, eval_func=get_sig)
        child = walker.next()
        while child:
            child = walker.next()

        if node.builder:
            sigs.append(self.module.signature(node.builder_sig_adapter()))
        return self.module.collect(filter(lambda x: not x is None, sigs))

    def csig(self, node):
        """
        Generate a node's content signature, the digested signature
        of its content.

        node - the node
        returns - the content signature
        """
        if not node.use_signature:
            return None
        #XXX If configured, use the content signatures from the
        #XXX .sconsign file if the timestamps match.
        csig = node.get_csig()
        if not csig is None:
            return csig
        
        return self.module.signature(node)

    def get_signature(self, node):
        """
        Get the appropriate signature for a node.

        node - the node
        returns - the signature or None if the signature could not
        be computed.

        This method does not store the signature in the node and
        in the .sconsign file.
        """

        if not node.use_signature:
            # This node type doesn't use a signature (e.g. a
            # directory) so bail right away.
            return None
        elif node.builder:
            return self.bsig(node)
        elif not node.exists():
            return None
        else:
            return self.csig(node)

    def current(self, node, newsig):
        """
        Check if a signature is up to date with respect to a node.

        node - the node whose signature will be checked
        newsig - the (presumably current) signature of the file

        returns - 1 if the file is current with the specified signature,
        0 if it isn't
        """

        c = node.current()
        if not c is None:
            # The node itself has told us whether or not it's
            # current without checking the signature.  The
            # canonical uses here are a "0" return for a file
            # that doesn't exist, or a directory.
            return c

        oldtime, oldbsig, oldcsig = node.get_prevsiginfo()

        if not node.builder and node.get_timestamp() == oldtime:
            return 1
        
        return self.module.current(newsig, oldbsig)



"""engine.SCons.Action

XXX

"""

#
# Copyright (c) 2001, 2002 Steven Knight
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY
# KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

__revision__ = "src/engine/SCons/Action.py 0.D006 2002/03/28 02:47:47 software"

import os
import os.path
import re
import string
import sys

import SCons.Util

print_actions = 1;
execute_actions = 1;

exitvalmap = {
    2 : 127,
    13 : 126,
}

if os.name == 'posix':

    def defaultSpawn(cmd, args, env):
        pid = os.fork()
        if not pid:
            # Child process.
            exitval = 127
            args = [ 'sh', '-c' ] + \
                   [ string.join(map(lambda x: string.replace(str(x),
                                                              ' ',
                                                              r'\ '),
                                     args)) ]
            try:
                os.execvpe('sh', args, env)
            except OSError, e:
                exitval = exitvalmap[e[0]]
                sys.stderr.write("scons: %s: %s\n" % (cmd, e[1]))
            os._exit(exitval)
        else:
            # Parent process.
            pid, stat = os.waitpid(pid, 0)
            ret = stat >> 8
            return ret

elif os.name == 'nt':

    def pathsearch(cmd, env):
        # In order to deal with the fact that 1.5.2 doesn't have
        # os.spawnvpe(), roll our own PATH search.
        if os.path.isabs(cmd):
            if not os.path.exists(cmd):
                exts = env['PATHEXT']
                if not SCons.Util.is_List(exts):
                    exts = string.split(exts, os.pathsep)
                for e in exts:
                    f = cmd + e
                    if os.path.exists(f):
                        return f
            else:
                return cmd
        else:
            path = env['PATH']
            if not SCons.Util.is_List(path):
                path = string.split(path, os.pathsep)
            exts = env['PATHEXT']
            if not SCons.Util.is_List(exts):
                exts = string.split(exts, os.pathsep)
            pairs = []
            for dir in path:
                for e in exts:
                    pairs.append((dir, e))
            for dir, ext in pairs:
                f = os.path.join(dir, cmd)
                if not ext is None:
                    f = f + ext
                if os.path.exists(f):
                    return f
        return None

    # Attempt to find cmd.exe (for WinNT/2k/XP) or
    # command.com for Win9x

    cmd_interp = ''
    # First see if we can look in the registry...
    if SCons.Util.can_read_reg:
        try:
            # Look for Windows NT system root
            k=SCons.Util.RegOpenKeyEx(SCons.Util.hkey_mod.HKEY_LOCAL_MACHINE,
                                          'Software\\Microsoft\\Windows NT\\CurrentVersion')
            val, tok = SCons.Util.RegQueryValueEx(k, 'SystemRoot')
            cmd_interp = os.path.join(val, 'System32\\cmd.exe')
        except SCons.Util.RegError:
            try:
                # Okay, try the Windows 9x system root
                k=SCons.Util.RegOpenKeyEx(SCons.Util.hkey_mod.HKEY_LOCAL_MACHINE,
                                              'Software\\Microsoft\\Windows\\CurrentVersion')
                val, tok = SCons.Util.RegQueryValueEx(k, 'SystemRoot')
                cmd_interp = os.path.join(val, 'command.com')
            except:
                pass
    if not cmd_interp:
        cmd_interp = pathsearch('cmd', os.environ)
        if not cmd_interp:
            cmd_interp = pathsearch('command', os.environ)

    # The upshot of all this is that, if you are using Python 1.5.2,
    # you had better have cmd or command.com in your PATH when you run
    # scons.

    def defaultSpawn(cmd, args, env):
        if not cmd_interp:
            sys.stderr.write("scons: Could not find command interpreter, is it in your PATH?\n")
            return 127
        else:
            try:

                a = [ cmd_interp, '/C', args[0] ]
                for arg in args[1:]:
                    if ' ' in arg or '\t' in arg:
                        arg = '"' + arg + '"'
                    a.append(arg)
                ret = os.spawnve(os.P_WAIT, cmd_interp, a, env)
            except OSError, e:
                ret = exitvalmap[e[0]]
                sys.stderr.write("scons: %s: %s\n" % (cmd, e[1]))
            return ret
else:
    def defaultSpawn(cmd, args, env):
        sys.stderr.write("scons: Unknown os '%s', cannot spawn command interpreter.\n" % os.name)
        sys.stderr.write("scons: Set your command handler with SetCommandHandler().\n")
        return 127

spawn = defaultSpawn

def SetCommandHandler(func):
    global spawn
    spawn = func

def GetCommandHandler():
    global spawn
    return spawn

class CommandGenerator:
    """
    Wrappes a command generator function so the Action() factory
    function can tell a generator function from a function action.
    """
    def __init__(self, generator):
        self.generator = generator


def Action(act):
    """A factory for action objects."""
    if isinstance(act, ActionBase):
        return act
    elif isinstance(act, CommandGenerator):
        return CommandGeneratorAction(act.generator)
    elif callable(act):
        return FunctionAction(act)
    elif SCons.Util.is_String(act):
        return CommandAction(act)
    elif SCons.Util.is_List(act):
        return ListAction(act)
    else:
        return None

class ActionBase:
    """Base class for actions that create output objects."""
    def __cmp__(self, other):
        return cmp(self.__dict__, other.__dict__)

    def show(self, string):
        print string

    def subst_dict(self, **kw):
        """Create a dictionary for substitution of construction
        variables.

        This translates the following special arguments:

            env    - the construction environment itself,
                     the values of which (CC, CCFLAGS, etc.)
                     are copied straight into the dictionary

            target - the target (object or array of objects),
                     used to generate the TARGET and TARGETS
                     construction variables

            source - the source (object or array of objects),
                     used to generate the SOURCES construction
                     variable

        Any other keyword arguments are copied into the
        dictionary."""

        dict = {}
        if kw.has_key('env'):
            dict.update(kw['env'])
            del kw['env']

        try:
            cwd = kw['dir']
        except:
            cwd = None
        else:
            del kw['dir']

        if kw.has_key('target'):
            t = kw['target']
            del kw['target']
            if not SCons.Util.is_List(t):
                t = [t]
            try:
                cwd = t[0].cwd
            except AttributeError:
                pass
            dict['TARGETS'] = SCons.Util.PathList(map(os.path.normpath, map(str, t)))
            if dict['TARGETS']:
                dict['TARGET'] = dict['TARGETS'][0]

        if kw.has_key('source'):
            s = kw['source']
            del kw['source']
            if not SCons.Util.is_List(s):
                s = [s]
            dict['SOURCES'] = SCons.Util.PathList(map(os.path.normpath, map(str, s)))

        dict.update(kw)

        return dict

_rm = re.compile(r'\$[()]')
_remove = re.compile(r'\$\(([^\$]|\$[^\(])*?\$\)')

class CommandAction(ActionBase):
    """Class for command-execution actions."""
    def __init__(self, string):
        self.command = string

    def execute(self, **kw):
        dict = apply(self.subst_dict, (), kw)
        import SCons.Util
        cmd_list = SCons.Util.scons_subst_list(self.command, dict, {}, _rm)
        for cmd_line in cmd_list:
            if len(cmd_line):
                if print_actions:
                    cl = []
                    for arg in cmd_line:
                        if ' ' in arg or '\t' in arg:
                            arg = '"' + arg + '"'
                        cl.append(arg)
                    self.show(string.join(cl))
                if execute_actions:
                    try:
                        ENV = kw['env']['ENV']
                    except:
                        import SCons.Defaults
                        ENV = SCons.Defaults.ConstructionEnvironment['ENV']
                    ret = spawn(cmd_line[0], cmd_line, ENV)
                    if ret:
                        return ret
        return 0

    def _sig_dict(self, kw):
        """Supply a dictionary for use in computing signatures.

        For signature purposes, it doesn't matter what targets or
        sources we use, so long as we use the same ones every time
        so the signature stays the same.  We supply an array of two
        of each to allow for distinction between TARGET and TARGETS.
        """
        kw['target'] = ['__t1__', '__t2__']
        kw['source'] = ['__s1__', '__s2__']
        return apply(self.subst_dict, (), kw)

    def get_raw_contents(self, **kw):
        """Return the complete contents of this action's command line.
        """
        return SCons.Util.scons_subst(self.command, self._sig_dict(kw), {})

    def get_contents(self, **kw):
        """Return the signature contents of this action's command line.

        This strips $(-$) and everything in between the string,
        since those parts don't affect signatures.
        """
        return SCons.Util.scons_subst(self.command, self._sig_dict(kw), {}, _remove)

class CommandGeneratorAction(ActionBase):
    """Class for command-generator actions."""
    def __init__(self, generator):
        self.generator = generator

    def execute(self, **kw):
        # ensure that target is a list, to make it easier to write
        # generator functions:
        import SCons.Util
        if kw.has_key("target") and not SCons.Util.is_List(kw["target"]):
            kw["target"] = [kw["target"]]

        gen_list = apply(self.generator, (), kw)
        gen_list = map(lambda x: map(str, x), gen_list)

        # Do environment variable substitution on returned command list
        dict = apply(self.subst_dict, (), kw)
        cmd_list = [ ]
        for gen_line in gen_list:
            cmd_list.append([])
            curr_line = cmd_list[-1]
            for gen_arg in gen_line:
                arg_list = SCons.Util.scons_subst_list(gen_arg, dict, {})
                curr_line.extend(arg_list[0])
                if(len(arg_list) > 1):
                    cmd_list.extend(arg_list[1:])
                    curr_line = cmd_list[-1]
        
        for cmd_line in filter(lambda x: x, cmd_list):
            if print_actions:
                self.show(cmd_line)
            if execute_actions:
                try:
                    ENV = kw['env']['ENV']
                except:
                    import SCons.Defaults
                    ENV = SCons.Defaults.ConstructionEnvironment['ENV']
                ret = spawn(cmd_line[0], cmd_line, ENV)
                if ret:
                    return ret

        return 0

    def get_contents(self, **kw):
        """Return the signature contents of this action's command line.

        This strips $(-$) and everything in between the string,
        since those parts don't affect signatures.
        """
	kw['source'] = ["__s1__", "__s2__"]
	kw['target'] = ["__t1__", "__t2__"]
	cmd_list = apply(self.generator, (), kw)
	cmd_list = map(lambda x: map(str, x), cmd_list)
	cmd_list = map(lambda x: string.join(x,  "\0"), cmd_list)
	cmd_list = map(lambda x: _remove.sub('', x), cmd_list)
	cmd_list = map(lambda x: filter(lambda y: y, string.split(x, "\0")), cmd_list)
	return cmd_list

class FunctionAction(ActionBase):
    """Class for Python function actions."""
    def __init__(self, function):
        self.function = function

    def execute(self, **kw):
        # if print_actions:
        # XXX:  WHAT SHOULD WE PRINT HERE?
        if execute_actions:
            if kw.has_key('target') and not \
               SCons.Util.is_List(kw['target']):
                kw['target'] = [ kw['target'] ]
            if kw.has_key('source') and not \
               SCons.Util.is_List(kw['source']):
                kw['source'] = [ kw['source'] ]
            return apply(self.function, (), kw)

    def get_contents(self, **kw):
        """Return the signature contents of this callable action.

        By providing direct access to the code object of the
        function, Python makes this extremely easy.  Hooray!
        """
        #XXX DOES NOT ACCOUNT FOR CHANGES IN ENVIRONMENT VARIABLES
        #THE FUNCTION MAY USE
        try:
            # "self.function" is a function.
            code = self.function.func_code.co_code
        except:
            # "self.function" is a callable object.
            code = self.function.__call__.im_func.func_code.co_code
        return str(code)

class ListAction(ActionBase):
    """Class for lists of other actions."""
    def __init__(self, list):
        self.list = map(lambda x: Action(x), list)

    def execute(self, **kw):
        for l in self.list:
            r = apply(l.execute, (), kw)
            if r != 0:
                return r
        return 0

    def get_contents(self, **kw):
        """Return the signature contents of this action list.

        Simple concatenation of the signatures of the elements.
        """

        return reduce(lambda x, y: x + str(y.get_contents()), self.list, "")



#
# Copyright (c) 2001, 2002 Steven Knight
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY
# KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

__revision__ = "src/engine/SCons/Scanner/Prog.py 0.D006 2002/03/28 02:47:47 software"

import copy
import string

import SCons.Node
import SCons.Node.FS
import SCons.Scanner
import SCons.Util

class NullProgScanner:
    """A do-nothing ProgScanner for Environments that have no LIBS."""
    def scan(node, env, args = []):
        return []

null_scanner = NullProgScanner()

def ProgScan(fs = SCons.Node.FS.default_fs):
    """Return a prototype Scanner instance for scanning executable
    files for static-lib dependencies"""
    ps = ProgScanner(scan, "ProgScan")
    ps.fs = fs
    return ps

class ProgScanner(SCons.Scanner.Base):
    def __init__(self, *args, **kw):
        apply(SCons.Scanner.Base.__init__, (self,) + args, kw)
        self.hash = None
        self.pathscanners = {}

    def instance(self, env):
        """
        Return a unique instance of a Prog scanner object for a
        given environment.
        """
        try:
            libs = env.Dictionary('LIBS')
        except KeyError:
            # There are no LIBS in this environment, so just return the
            # fake "scanner" instance that always returns a null list.
            return null_scanner
        if SCons.Util.is_String(libs):
            libs = string.split(libs)

        try:
            dirs = tuple(SCons.Node.arg2nodes(env.Dictionary('LIBPATH'),
                                              self.fs.Dir))
        except:
            dirs = ()

        try:
            prefix = env.Dictionary('LIBPREFIX')
        except KeyError:
            prefix = ''

        try:
            suffix = env.Dictionary('LIBSUFFIX')
        except KeyError:
            suffix = ''

        key = (dirs, tuple(libs), prefix, suffix)
        if not self.pathscanners.has_key(key):
            clone = copy.copy(self)
            clone.hash = key
            clone.argument = [self.fs, dirs, libs, prefix, suffix]    # XXX reaching into object
            self.pathscanners[key] = clone
        return self.pathscanners[key]

    def __hash__(self):
        return hash(self.hash)

def scan(node, env, args = [SCons.Node.FS.default_fs, (), [], '', '']):
    """
    This scanner scans program files for static-library
    dependencies.  It will search the LIBPATH environment variable
    for libraries specified in the LIBS variable, returning any
    files it finds as dependencies.
    """
    fs, libpath, libs, prefix, suffix = args
    libs = map(lambda x, s=suffix, p=prefix: p + x + s, libs)
    return SCons.Node.FS.find_files(libs, libpath, fs.File)



"""SCons.Sig.TimeStamp

The TimeStamp signature package for the SCons software construction
utility.

"""

#
# Copyright (c) 2001, 2002 Steven Knight
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY
# KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

__revision__ = "src/engine/SCons/Sig/TimeStamp.py 0.D006 2002/03/28 02:47:47 software"

def current(new, old):
    """Return whether a new timestamp is up-to-date with
    respect to an old timestamp.
    """
    return not old is None and new <= old

def collect(signatures):
    """
    Collect a list of timestamps, returning
    the most-recent timestamp from the list 

    signatures - a list of timestamps
    returns - the most recent timestamp
    """

    if len(signatures) == 0:
        return 0
    elif len(signatures) == 1:
        return signatures[0]
    else:
        return max(signatures)

def signature(obj):
    """Generate a timestamp.
    """
    return obj.get_timestamp()

def to_string(signature):
    """Convert a timestamp to a string"""
    return str(signature)

def from_string(string):
    """Convert a string to a timestamp"""
    return int(string)





"""SCons.Taskmaster

Generic Taskmaster.

"""

#
# Copyright (c) 2001, 2002 Steven Knight
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY
# KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

__revision__ = "src/engine/SCons/Taskmaster.py 0.D006 2002/03/28 02:47:47 software"




import SCons.Node
import string
import SCons.Errors
import copy

class Task:
    """Default SCons build engine task.
    
    This controls the interaction of the actual building of node
    and the rest of the engine.

    This is expected to handle all of the normally-customizable
    aspects of controlling a build, so any given application
    *should* be able to do what it wants by sub-classing this
    class and overriding methods as appropriate.  If an application
    needs to customze something by sub-classing Taskmaster (or
    some other build engine class), we should first try to migrate
    that functionality into this class.
    
    Note that it's generally a good idea for sub-classes to call
    these methods explicitly to update state, etc., rather than
    roll their own interaction with Taskmaster from scratch."""
    def __init__(self, tm, targets, top, scanner = None):
        self.tm = tm
        self.targets = targets
        self.top = top
        self.scanner = scanner

    def execute(self):
        if self.targets[0].get_state() != SCons.Node.up_to_date:
            self.targets[0].prepare()
            self.targets[0].build()

    def get_target(self):
        """Fetch the target being built or updated by this task.
        """
        return self.targets[0]

    def set_tstates(self, state):
        """Set all of the target nodes's states."""
        for t in self.targets:
            t.set_state(state)

    def executed(self):
        """Called when the task has been successfully executed.

        This may have been a do-nothing operation (to preserve
        build order), so check the node's state before updating
        things.  Most importantly, this calls back to the
        Taskmaster to put any node tasks waiting on this one
        back on the pending list."""
        if self.targets[0].get_state() == SCons.Node.executing:
            self.set_tstates(SCons.Node.executed)
            for t in self.targets:
                t.store_sigs()
            parents = {}
            for p in reduce(lambda x, y: x + y.get_parents(), self.targets, []):
                parents[p] = 1
            ready = filter(lambda x, s=self.scanner:
                                  (x.get_state() == SCons.Node.pending
                                   and x.children_are_executed(s)),
                           parents.keys())
            tasks = {}
            for t in map(lambda r: r.task, ready):
                tasks[t] = 1
            self.tm.pending_to_ready(tasks.keys())

    def failed(self):
        """Default action when a task fails:  stop the build."""
        self.fail_stop()

    def fail_stop(self):
        """Explicit stop-the-build failure."""
        self.set_tstates(SCons.Node.failed)
        self.tm.stop()

    def fail_continue(self):
        """Explicit continue-the-build failure.

        This sets failure status on the target nodes and all of
        their dependent parent nodes.
        """
        nodes = {}
        for t in self.targets:
            def get_parents(node, parent): return node.get_parents()
	    def set_nodes(node, parent, nodes=nodes): nodes[node] = 1
            walker = SCons.Node.Walker(t, get_parents, eval_func=set_nodes)
            n = walker.next()
            while n:
                n = walker.next()
        pending = filter(lambda x: x.get_state() == SCons.Node.pending,
                         nodes.keys())
        tasks = {}
        for t in map(lambda r: r.task, pending):
            tasks[t] = 1
        self.tm.pending_remove(tasks.keys())

    def make_ready(self):
        """Make a task ready for execution."""
        state = SCons.Node.up_to_date
        for t in self.targets:
            bsig = self.tm.calc.bsig(t)
            t.set_bsig(bsig)
            if not self.tm.calc.current(t, bsig):
                state = SCons.Node.executing
        self.set_tstates(state)
        self.tm.add_ready(self)



class Calc:
    def bsig(self, node):
        """
        """
        return None

    def current(self, node, sig):
        """Default SCons build engine is-it-current function.
    
        This returns "always out of date," so every node is always
        built/visited.
        """
        return 0



class Taskmaster:
    """A generic Taskmaster for handling a bunch of targets.

    Classes that override methods of this class should call
    the base class method, so this class can do its thing.    
    """

    def __init__(self, targets=[], tasker=Task, calc=Calc()):
        
        def out_of_date(node, parent):
            if node.get_state():
                # The state is set, so someone has already been here
                # (finished or currently executing).  Find another one.
                return []
            # Scan the file before fetching its children().
            if parent:
                scanner = parent.src_scanner_get(node.scanner_key())
            else:
                scanner = None
            return filter(lambda x: x.get_state() != SCons.Node.up_to_date,
                          node.children(scanner))

        def cycle_error(node, stack):
            if node.builder:
                nodes = stack + [node]
                nodes.reverse()
                desc = "Dependency cycle: " + string.join(map(str, nodes), " -> ")
                raise SCons.Errors.UserError, desc

        def eval_node(node, parent, self=self):
            if node.get_state():
                # The state is set, so someone has already been here
                # (finished or currently executing).  Find another one.
                return
            if not node.builder:
                # It's a source file, we don't need to build it,
                # but mark it as "up to date" so targets won't
                # wait for it.
                node.set_state(SCons.Node.up_to_date)
                # set the signature for non-derived files
                # here so they don't get recalculated over
                # and over again:
                node.set_csig(self.calc.csig(node))
                return
            try:
                tlist = node.builder.targets(node)
            except AttributeError:
                tlist = [ node ]
            if parent:
                scanner = parent.src_scanner_get(node.scanner_key())
            else:
                scanner = None
            task = self.tasker(self, tlist, self.walkers[0].is_done(), scanner)
            if not tlist[0].children_are_executed(scanner):
                for t in tlist:
                    t.set_state(SCons.Node.pending)
                    t.task = task
                self.pending = self.pending + 1
                return
            task.make_ready()

        #XXX In Python 2.2 we can get rid of f1, f2 and f3:
        self.walkers = map(lambda x, f1=out_of_date,
                                     f2=cycle_error,
                                     f3=eval_node:
                                  SCons.Node.Walker(x, f1, f2, f3),
                           targets)
        self.tasker = tasker
        self.calc = calc
        self.ready = []
        self.pending = 0
        
        self._find_next_ready_node()

    def next_task(self):
        """Return the next task to be executed."""
        if self.ready:
            task = self.ready.pop()
            if not self.ready:
                self._find_next_ready_node()
            return task
        else:
            return None

    def _find_next_ready_node(self):
        """Find the next node that is ready to be built"""
        while self.walkers:
            n = self.walkers[0].next()
            if n == None:
                self.walkers.pop(0)
                continue
            if self.ready:
                return
            
    def is_blocked(self):
        return not self.ready and self.pending

    def stop(self):
        """Stop the current build completely."""
        self.walkers = []
        self.pending = 0
        self.ready = []

    def add_ready(self, task):
        """Add a task to the ready queue.
        """
        self.ready.append(task)

    def pending_to_ready(self, tasks):
        """Move the specified tasks from the pending count
        to the 'ready' queue.
        """
        self.pending_remove(tasks)
        for t in tasks:
            t.make_ready()

    def pending_remove(self, tasks):
        """Remove tasks from the pending count.
        
        We assume that the caller has already confirmed that
        the nodes in this task are in pending state.
        """
        self.pending = self.pending - len(tasks)




"""scons.Node.Alias

Alias nodes.

This creates a hash of global Aliases (dummy targets).

"""

#
# Copyright (c) 2001, 2002 Steven Knight
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY
# KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

__revision__ = "src/engine/SCons/Node/Alias.py 0.D006 2002/03/28 02:47:47 __USER__"

import UserDict

import SCons.Errors
import SCons.Node
import SCons.Util

class AliasNameSpace(UserDict.UserDict):
    def Alias(self, name):
        if self.has_key(name):
            raise SCons.Errors.UserError
        self[name] = SCons.Node.Alias.Alias(name)
        return self[name]

    def lookup(self, name):
        try:
            return self[name]
        except KeyError:
            return None

class Alias(SCons.Node.Node):
    def __init__(self, name):
        SCons.Node.Node.__init__(self)
        self.name = name

    def __str__(self):
        return self.name

    def build(self):
        """A "builder" for aliases."""
        pass

    def set_bsig(self, bsig):
        """An alias has no signature."""
        pass

    def set_csig(self, csig):
        """An alias has no signature."""
        pass

    def current(self):
        """If all of our children were up-to-date, then this
        Alias was up-to-date, too."""
        state = 0
        for kid in self.children(None):
            s = kid.get_state()
            if s and (not state or s > state):
                state = s
        if state == 0 or state == SCons.Node.up_to_date:
            return 1
        else:
            return 0

    def sconsign(self):
        """An Alias is not recorded in .sconsign files"""
        pass
        
default_ans = AliasNameSpace()

SCons.Node.arg2nodes_lookups.append(default_ans.lookup)



"""SCons.Environment

XXX

"""

#
# Copyright (c) 2001, 2002 Steven Knight
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY
# KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

__revision__ = "src/engine/SCons/Environment.py 0.D006 2002/03/28 02:47:47 software"


import copy
import os
import os.path
import re
import shutil
import string
import sys
import types

import SCons.Builder
import SCons.Defaults
from SCons.Errors import UserError
import SCons.Node
import SCons.Node.FS
import SCons.Util

def installFunc(env, target, source):
    try:
        map(lambda t: os.unlink(str(t)), target)
    except OSError:
        pass
    
    try:
        SCons.Node.FS.file_link(str(source[0]), str(target[0]))
        print 'Install file: "%s" as "%s"' % \
              (source[0], target[0])
        return 0
    except IOError, e:
        sys.stderr.write('Unable to install "%s" as "%s"\n%s\n' % \
                         (source[0], target[0], str(e)))
        return -1
    except OSError, e:
        sys.stderr.write('Unable to install "%s" as "%s"\n%s\n' % \
                         (source[0], target[0], str(e)))
        return -1

InstallBuilder = SCons.Builder.Builder(name='Install',
                                       action=installFunc)

def our_deepcopy(x):
   """deepcopy lists and dictionaries, and just copy the reference 
   for everything else.""" 
   if SCons.Util.is_Dict(x):
       copy = {}
       for key in x.keys():
           copy[key] = our_deepcopy(x[key])
   elif SCons.Util.is_List(x):
       copy = map(our_deepcopy, x)
   else:
       copy = x
   return copy

class Environment:
    """Base class for construction Environments.  These are
    the primary objects used to communicate dependency and
    construction information to the build engine.

    Keyword arguments supplied when the construction Environment
    is created are construction variables used to initialize the
    Environment.
    """

    def __init__(self, **kw):
        self.fs = SCons.Node.FS.default_fs
        self._dict = our_deepcopy(SCons.Defaults.ConstructionEnvironment)
        apply(self.Update, (), kw)

        #
        # self.autogen_vars is a tuple of tuples.  Each inner tuple
        # has four elements, each strings referring to an environment
        # variable, and describing how to autogenerate a particular
        # variable.  The elements are:
        #
        # 0 - The variable to generate
        # 1 - The "source" variable, usually a list
        # 2 - The "prefix" variable
        # 3 - The "suffix" variable
        #
        # The autogenerated variable is a list, consisting of every
        # element of the source list, or a single element if the source
        # is a string, with the prefix and suffix concatenated.
        #
        self.autogen_vars = ( VarInterpolator('_LIBFLAGS',
                                              'LIBS',
                                              'LIBLINKPREFIX',
                                              'LIBLINKSUFFIX'),
                              DirVarInterp('_LIBDIRFLAGS',
                                           'LIBPATH',
                                           'LIBDIRPREFIX',
                                           'LIBDIRSUFFIX' ),
                              DirVarInterp('_INCFLAGS',
                                           'CPPPATH',
                                           'INCPREFIX',
                                           'INCSUFFIX') )

    def __cmp__(self, other):
	return cmp(self._dict, other._dict)

    def Builders(self):
	pass	# XXX

    def Copy(self, **kw):
	"""Return a copy of a construction Environment.  The
	copy is like a Python "deep copy"--that is, independent
	copies are made recursively of each objects--except that
	a reference is copied when an object is not deep-copyable
	(like a function).  There are no references to any mutable
	objects in the original Environment.
	"""
        clone = copy.copy(self)
        clone._dict = our_deepcopy(self._dict)
	apply(clone.Update, (), kw)
	return clone

    def Scanners(self):
	pass	# XXX

    def	Update(self, **kw):
	"""Update an existing construction Environment with new
	construction variables and/or values.
	"""
	self._dict.update(our_deepcopy(kw))
        if self._dict.has_key('BUILDERS') and \
           not SCons.Util.is_List(self._dict['BUILDERS']):
            self._dict['BUILDERS'] = [self._dict['BUILDERS']]
        if self._dict.has_key('SCANNERS') and \
           not SCons.Util.is_List(self._dict['SCANNERS']):
            self._dict['SCANNERS'] = [self._dict['SCANNERS']]

        class BuilderWrapper:
            """Wrapper class that allows an environment to
            be associated with a Builder at instantiation.
            """
            def __init__(self, env, builder):
                self.env = env
                self.builder = builder

            def __call__(self, target = None, source = None):
                return self.builder(self.env, target, source)

            # This allows a Builder to be executed directly
            # through the Environment to which it's attached.
            # In practice, we shouldn't need this, because
            # builders actually get executed through a Node.
            # But we do have a unit test for this, and can't
            # yet rule out that it would be useful in the
            # future, so leave it for now.
            def execute(self, **kw):
                kw['env'] = self.env
                apply(self.builder.execute, (), kw)

        for b in self._dict['BUILDERS']:
            setattr(self, b.name, BuilderWrapper(self, b))

        for s in self._dict['SCANNERS']:
            setattr(self, s.name, s)

    def	Depends(self, target, dependency):
	"""Explicity specify that 'target's depend on 'dependency'."""
        tlist = SCons.Node.arg2nodes(target, self.fs.File)
        dlist = SCons.Node.arg2nodes(dependency, self.fs.File)
	for t in tlist:
	    t.add_dependency(dlist)

	if len(tlist) == 1:
	    tlist = tlist[0]
	return tlist

    def Ignore(self, target, dependency):
        """Ignore a dependency."""
        tlist = SCons.Node.arg2nodes(target, self.fs.File)
        dlist = SCons.Node.arg2nodes(dependency, self.fs.File)
        for t in tlist:
            t.add_ignore(dlist)

        if len(tlist) == 1:
            tlist = tlist[0]
        return tlist

    def Precious(self, *targets):
        tlist = []
        for t in targets:
            tlist.extend(SCons.Node.arg2nodes(t, self.fs.File))

        for t in tlist:
            t.set_precious()

        if len(tlist) == 1:
            tlist = tlist[0]
        return tlist

    def Dictionary(self, *args):
	if not args:
	    return self._dict
	dlist = map(lambda x, s=self: s._dict[x], args)
	if len(dlist) == 1:
	    dlist = dlist[0]
	return dlist

    def __setitem__(self, key, value):
        self._dict[key] = value

    def __getitem__(self, key):
        return self._dict[key]

    def __delitem__(self, key):
        del self._dict[key]

    def Command(self, target, source, action):
        """Builds the supplied target files from the supplied
        source files using the supplied action.  Action may
        be any type that the Builder constructor will accept
        for an action."""
        bld = SCons.Builder.Builder(name="Command", action=action)
        return bld(self, target, source)

    def Install(self, dir, source):
        """Install specified files in the given directory."""
        sources = SCons.Node.arg2nodes(source, self.fs.File)
        dnodes = SCons.Node.arg2nodes(dir, self.fs.Dir)
        tgt = []
        for dnode in dnodes:
            for src in sources:
                target = SCons.Node.FS.default_fs.File(src.name, dnode)
                tgt.append(InstallBuilder(self, target, src))
        if len(tgt) == 1:
            tgt = tgt[0]
        return tgt

    def InstallAs(self, target, source):
        """Install sources as targets."""
        sources = SCons.Node.arg2nodes(source, self.fs.File)
        targets = SCons.Node.arg2nodes(target, self.fs.File)
        ret = []
        for src, tgt in map(lambda x, y: (x, y), sources, targets):
            ret.append(InstallBuilder(self, tgt, src))
        if len(ret) == 1:
            ret = ret[0]
        return ret
  
    def subst(self, string):
	"""Recursively interpolates construction variables from the
	Environment into the specified string, returning the expanded
	result.  Construction variables are specified by a $ prefix
	in the string and begin with an initial underscore or
	alphabetic character followed by any number of underscores
	or alphanumeric characters.  The construction variable names
	may be surrounded by curly braces to separate the name from
	trailing characters.
	"""
	return SCons.Util.scons_subst(string, self._dict, {})

    def get_scanner(self, skey):
        """Find the appropriate scanner given a key (usually a file suffix).
        Does a linear search. Could be sped up by creating a dictionary if
        this proves too slow.
        """
        if self._dict['SCANNERS']:
            for scanner in self._dict['SCANNERS']:
                if skey in scanner.skeys:
                    return scanner
        return None

    def autogenerate(self, fs = SCons.Node.FS.default_fs, dir = None):
        """Return a dictionary of autogenerated "interpolated"
        construction variables.
        """
        dict = {}
        for interp in self.autogen_vars:
            interp.instance(dir, fs).generate(dict, self._dict)
        return dict

class VarInterpolator:
    def __init__(self, dest, src, prefix, suffix):
        self.dest = dest
        self.src = src
        self.prefix = prefix
        self.suffix = suffix

    def prepareSrc(self, dict):
        src = dict[self.src]
	if SCons.Util.is_String(src):
	    src = string.split(src)
        elif not SCons.Util.is_List(src):
            src = [ src ]

        def prepare(x, dict=dict):
            if isinstance(x, SCons.Node.Node):
                return x
            else:
                return SCons.Util.scons_subst(x, {}, dict)

        return map(prepare, src)

    def generate(self, ddict, sdict):
        if not sdict.has_key(self.src):
            ddict[self.dest] = ''
            return

        src = filter(lambda x: not x is None, self.prepareSrc(sdict))

        if not src:
            ddict[self.dest] = ''
            return

        prefix = sdict.get(self.prefix, '')
        suffix = sdict.get(self.suffix, '')

        def autogenFunc(x, suff=suffix, pref=prefix):
            """Generate the interpolated variable.  If the prefix
            ends in a space, or the suffix begins in a space,
            leave it as a separate element of the list."""
            ret = [ str(x) ]
            if pref and pref[-1] == ' ':
                ret.insert(0, pref[:-1])
            else:
                ret[0] = pref + ret[0]
            if suff and suff[0] == ' ':
                ret.append(suff[1:])
            else:
                ret[-1] = ret[-1] + suff
            return ret

        ddict[self.dest] = reduce(lambda x, y: x+y,
                                  map(autogenFunc, src))

    def instance(self, dir, fs):
        return self

class DirVarInterp(VarInterpolator):
    def __init__(self, dest, src, prefix, suffix):
        VarInterpolator.__init__(self, dest, src, prefix, suffix)
        self.fs = None
        self.Dir = None
        self.dictInstCache = {}

    def prepareSrc(self, dict):
        src = VarInterpolator.prepareSrc(self, dict)

        def prepare(x, self=self):
            if isinstance(x, SCons.Node.Node):
                return x
            elif str(x):
                return self.fs.Dir(str(x), directory=self.dir)
            else:
                return None

        return map(prepare, src)

    def instance(self, dir, fs):
        try:
            ret = self.dictInstCache[(dir, fs)]
        except KeyError:
            ret = copy.copy(self)
            ret.fs = fs
            ret.dir = dir
            self.dictInstCache[(dir, fs)] = ret
        return ret

    def generate(self, ddict, sdict):
        VarInterpolator.generate(self, ddict, sdict)
        if ddict[self.dest]:
            ddict[self.dest] = ['$('] + ddict[self.dest] + ['$)']



"""SCons.Util

Various utility functions go here.

"""

#
# Copyright (c) 2001, 2002 Steven Knight
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY
# KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

__revision__ = "src/engine/SCons/Util.py 0.D006 2002/03/28 02:47:47 software"


import copy
import os
import os.path
import re
import stat
import string
import sys
import types
import UserDict
import UserList

try:
    from UserString import UserString
except ImportError:
    class UserString:
        pass



class PathList(UserList.UserList):
    """This class emulates the behavior of a list, but also implements
    the special "path dissection" attributes we can use to find
    suffixes, base names, etc. of the paths in the list.

    One other special attribute of this class is that, by
    overriding the __str__ and __repr__ methods, this class
    represents itself as a space-concatenated string of
    the list elements, as in:

    >>> pl=PathList(["/foo/bar.txt", "/baz/foo.txt"])
    >>> pl
    '/foo/bar.txt /baz/foo.txt'
    >>> pl.base
    'bar foo'
    """
    def __init__(self, seq = []):
        UserList.UserList.__init__(self, seq)

    def __getattr__(self, name):
        # This is how we implement the "special" attributes
        # such as base, suffix, basepath, etc.
        try:
            return self.dictSpecialAttrs[name](self)
        except KeyError:
            raise AttributeError, 'PathList has no attribute: %s' % name

    def __splitPath(self, split_func=os.path.split):
        """This method calls the supplied split_func on each element
        in the contained list.  We expect split_func to return a
        2-tuple, usually representing two elements of a split file path,
        such as those returned by os.path.split().

        We return a 2-tuple of lists, each equal in length to the contained
        list.  The first list represents all the elements from the
        first part of the split operation, the second represents
        all elements from the second part."""
        list1 = []
        list2 = []
        for strPath in self.data:
            first_part, second_part = split_func(strPath)
            list1.append(first_part)
            list2.append(second_part)
        return (self.__class__(list1),
                self.__class__(list2))

    def __getBasePath(self):
        """Return the file's directory and file name, with the
        suffix stripped."""
        return self.__splitPath(os.path.splitext)[0]

    def __getSuffix(self):
        """Return the file's suffix."""
        return self.__splitPath(os.path.splitext)[1]

    def __getFileName(self):
        """Return the file's name without the path."""
        return self.__splitPath()[1]

    def __getDir(self):
        """Return the file's path."""
        return self.__splitPath()[0]

    def __getBase(self):
        """Return the file name with path and suffix stripped."""
        return self.__getFileName().__splitPath(os.path.splitext)[0]

    dictSpecialAttrs = { "file" : __getFileName,
                         "base" : __getBasePath,
                         "filebase" : __getBase,
                         "dir" : __getDir,
                         "suffix" : __getSuffix }
    
    def __str__(self):
        return string.join(self.data)

    def __repr__(self):
        return repr(string.join(self.data))

    def __getitem__(self, item):
        # We must do this to ensure that single items returned
        # by index access have the special attributes such as
        # suffix and basepath.
        return self.__class__([ UserList.UserList.__getitem__(self, item), ])

_cv = re.compile(r'\$([_a-zA-Z]\w*|{[^}]*})')
_space_sep = re.compile(r'[\t ]+(?![^{]*})')

def scons_subst_list(strSubst, globals, locals, remove=None):
    """
    This function is similar to scons_subst(), but with
    one important difference.  Instead of returning a single
    string, this function returns a list of lists.
    The first (outer) list is a list of lines, where the
    substituted stirng has been broken along newline characters.
    The inner lists are lists of command line arguments, i.e.,
    the argv array that should be passed to a spawn or exec
    function.

    One important thing this guy does is preserve environment
    variables that are lists.  For instance, if you have
    an environment variable that is a Python list (or UserList-
    derived class) that contains path names with spaces in them,
    then the entire path will be returned as a single argument.
    This is the only way to know where the 'split' between arguments
    is for executing a command line."""

    def repl(m, globals=globals, locals=locals):
        key = m.group(1)
        if key[:1] == '{' and key[-1:] == '}':
            key = key[1:-1]
	try:
            e = eval(key, globals, locals)
            if e is None:
                s = ''
            elif is_List(e):
                s = string.join(map(str, e), '\0')
            else:
                s = _space_sep.sub('\0', str(e))
	except NameError:
	    s = ''
	return s
    n = 1

    # Tokenize the original string...
    strSubst = _space_sep.sub('\0', str(strSubst))
    
    # Now, do the substitution
    while n != 0:
        strSubst, n = _cv.subn(repl, strSubst)
    # Now parse the whole list into tokens.
    listLines = string.split(strSubst, '\n')
    if remove:
        listLines = map(lambda x,re=remove: re.sub('', x), listLines)
    return map(lambda x: filter(lambda y: y, string.split(x, '\0')),
               listLines)

def scons_subst(strSubst, globals, locals, remove=None):
    """Recursively interpolates dictionary variables into
    the specified string, returning the expanded result.
    Variables are specified by a $ prefix in the string and
    begin with an initial underscore or alphabetic character
    followed by any number of underscores or alphanumeric
    characters.  The construction variable names may be
    surrounded by curly braces to separate the name from
    trailing characters.
    """
    cmd_list = scons_subst_list(strSubst, globals, locals, remove)
    return string.join(map(string.join, cmd_list), '\n')

def render_tree(root, child_func, margin=[0], visited={}):
    """
    Render a tree of nodes into an ASCII tree view.
    root - the root node of the tree
    child_func - the function called to get the children of a node
    margin - the format of the left margin to use for children of root.
       1 results in a pipe, and 0 results in no pipe.
    visited - a dictionart of visited nodes in the current branch
    """

    if visited.has_key(root):
        return ""

    children = child_func(root)
    retval = ""
    for pipe in margin[:-1]:
        if pipe:
            retval = retval + "| "
        else:
            retval = retval + "  "

    retval = retval + "+-" + str(root) + "\n"
    visited = copy.copy(visited)
    visited[root] = 1

    for i in range(len(children)):
        margin.append(i<len(children)-1)
        retval = retval + render_tree(children[i], child_func, margin, visited
)
        margin.pop()

    return retval

def is_Dict(e):
    return type(e) is types.DictType or isinstance(e, UserDict.UserDict)

def is_List(e):
    return type(e) is types.ListType or isinstance(e, UserList.UserList)

if hasattr(types, 'UnicodeType'):
    def is_String(e):
        return type(e) is types.StringType \
            or type(e) is types.UnicodeType \
            or isinstance(e, UserString)
else:
    def is_String(e):
        return type(e) is types.StringType or isinstance(e, UserString)

# attempt to load the windows registry module:
can_read_reg = 0
try:
    import _winreg

    can_read_reg = 1
    hkey_mod = _winreg

    RegOpenKeyEx = _winreg.OpenKeyEx
    RegEnumKey = _winreg.EnumKey
    RegEnumValue = _winreg.EnumValue
    RegQueryValueEx = _winreg.QueryValueEx
    RegError = _winreg.error

except ImportError:
    try:
        import win32api
        import win32con
        can_read_reg = 1
        hkey_mod = win32con

        RegOpenKeyEx = win32api.RegOpenKeyEx
        RegEnumKey = win32api.RegEnumKey
        RegEnumValue = win32api.RegEnumValue
        RegQueryValueEx = win32api.RegQueryValueEx
        RegError = win32api.error

    except ImportError:
        pass

if can_read_reg:
    HKEY_CLASSES_ROOT = hkey_mod.HKEY_CLASSES_ROOT
    HKEY_LOCAL_MACHINE = hkey_mod.HKEY_LOCAL_MACHINE
    HKEY_CURRENT_USER = hkey_mod.HKEY_CURRENT_USER
    HKEY_USERS = hkey_mod.HKEY_USERS


if sys.platform == 'win32':

    def WhereIs(file, path=None, pathext=None):
        if path is None:
            path = os.environ['PATH']
        if is_String(path):
            path = string.split(path, os.pathsep)
        if pathext is None:
            pathext = os.environ['PATHEXT']
        if is_String(pathext):
            pathext = string.split(pathext, os.pathsep)
        for ext in pathext:
            if string.lower(ext) == string.lower(file[-len(ext):]):
                pathext = ['']
                break
        for dir in path:
            f = os.path.join(dir, file)
            for ext in pathext:
                fext = f + ext
                if os.path.isfile(fext):
                    return fext
        return None

else:

    def WhereIs(file, path=None, pathext=None):
        if path is None:
            path = os.environ['PATH']
        if is_String(path):
            path = string.split(path, os.pathsep)
        for dir in path:
            f = os.path.join(dir, file)
            if os.path.isfile(f):
                try:
                    st = os.stat(f)
                except:
                    continue
                if stat.S_IMODE(st[stat.ST_MODE]) & 0111:
                    return f
        return None



"""engine.SCons.script

This file implements the main() function used by the scons script.

Architecturally, this *is* the scons script, and will likely only be
called from the external "scons" wrapper.  Consequently, anything here
should not be, or be considered, part of the build engine.  If it's
something that we expect other software to want to use, it should go in
some other module.  If it's specific to the "scons" script invocation,
it goes here.

"""

#
# Copyright (c) 2001, 2002 Steven Knight
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY
# KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

__revision__ = "src/engine/SCons/Script/__init__.py 0.D006 2002/03/28 02:47:47 software"

import getopt
import os
import os.path
import string
import sys
import traceback
import copy

# Strip the script directory from sys.path() so on case-insensitive
# (WIN32) systems Python doesn't think that the "scons" script is the
# "SCons" package.  Replace it with our own version directory so, if
# if they're there, we pick up the right version of the build engine
# modules.
sys.path = [os.path.join(sys.prefix, 'lib', 'scons-0.06')] + sys.path[1:]

import SCons.Node
import SCons.Node.FS
import SCons.Job
from SCons.Errors import *
import SCons.Sig
import SCons.Sig.MD5
from SCons.Taskmaster import Taskmaster
import SCons.Builder
import SCons.Script.SConscript


#
# Task control.
#
class BuildTask(SCons.Taskmaster.Task):
    """An SCons build task."""
    def execute(self):
        if self.targets[0].get_state() == SCons.Node.up_to_date:
            if self.top:
                print 'scons: "%s" is up to date.' % str(self.targets[0])
        else:
            try:
                self.targets[0].prepare()
                self.targets[0].build()
            except BuildError, e:
                sys.stderr.write("scons: *** [%s] %s\n" % (e.node, e.errstr))
                if e.errstr == 'Exception':
                    traceback.print_exception(e.args[0], e.args[1],
                                              e.args[2])
                raise

    def executed(self):
        SCons.Taskmaster.Task.executed(self)
        # print the tree here instead of in execute() because
        # this method is serialized, but execute isn't:
        if print_tree and self.top:
            print
            print SCons.Util.render_tree(self.targets[0], get_children)

    def failed(self):
        global exit_status
        if ignore_errors:
            SCons.Taskmaster.Task.executed(self)
        elif keep_going_on_error:
            SCons.Taskmaster.Task.fail_continue(self)
            exit_status = 2
        else:
            SCons.Taskmaster.Task.fail_stop(self)
            exit_status = 2

class CleanTask(SCons.Taskmaster.Task):
    """An SCons clean task."""
    def execute(self):
        if self.targets[0].builder:
            try:
                os.unlink(self.targets[0].path)
            except OSError:
                pass
            else:
                print "Removed " + self.targets[0].path
            try:
                for t in self.targets[1:]:
                    try:
                        os.unlink(t.path)
                    except OSError:
                        pass
            except IndexError:
                pass

class QuestionTask(SCons.Taskmaster.Task):
    """An SCons task for the -q (question) option."""
    def execute(self):
        if self.targets[0].get_state() != SCons.Node.up_to_date:
            global exit_status
            exit_status = 1
            self.tm.stop()

    def executed(self):
        pass

# Global variables

include_dirs = []
num_jobs = 1
scripts = []
task_class = BuildTask	# default action is to build targets
current_func = None
calc = None
ignore_errors = 0
keep_going_on_error = 0
help_option = None
print_tree = 0
climb_up = 0
target_top = None
exit_status = 0 # exit status, assume success by default

# utility functions

def get_children(node): return node.all_children(None)

def _scons_syntax_error(e):
    """Handle syntax errors. Print out a message and show where the error
    occurred.
    """
    etype, value, tb = sys.exc_info()
    lines = traceback.format_exception_only(etype, value)
    for line in lines:
        sys.stderr.write(line+'\n')
    sys.exit(2)

def _scons_user_error(e):
    """Handle user errors. Print out a message and a description of the
    error, along with the line number and routine where it occured.
    """
    etype, value, tb = sys.exc_info()
    while tb.tb_next is not None:
        tb = tb.tb_next
    lineno = traceback.tb_lineno(tb)
    filename = tb.tb_frame.f_code.co_filename
    routine = tb.tb_frame.f_code.co_name
    sys.stderr.write("\nSCons error: %s\n" % value)
    sys.stderr.write('File "%s", line %d, in %s\n' % (filename, lineno, routine))
    sys.exit(2)

def _scons_user_warning(e):
    """Handle user warnings. Print out a message and a description of
    the warning, along with the line number and routine where it occured.
    """
    etype, value, tb = sys.exc_info()
    while tb.tb_next is not None:
        tb = tb.tb_next
    lineno = traceback.tb_lineno(tb)
    filename = tb.tb_frame.f_code.co_filename
    routine = tb.tb_frame.f_code.co_name
    sys.stderr.write("\nSCons warning: %s\n" % e)
    sys.stderr.write('File "%s", line %d, in %s\n' % (filename, lineno, routine))

def _scons_other_errors():
    """Handle all errors but user errors. Print out a message telling
    the user what to do in this case and print a normal trace.
    """
    print 'other errors'
    traceback.print_exc()
    sys.exit(2)


#
# After options are initialized, the following variables are
# filled in:
#
option_list = []	# list of Option objects
short_opts = ""		# string of short (single-character) options
long_opts = []		# array of long (--) options
opt_func = {}		# mapping of option strings to functions

def options_init():
    """Initialize command-line options processing.
    
    This is in a subroutine mainly so we can easily single-step over
    it in the debugger.
    """

    class Option:
	"""Class for command-line option information.

	This exists to provide a central location for everything
	describing a command-line option, so that we can change
	options without having to update the code to handle the
	option in one place, the -h help message in another place,
	etc.  There are no methods here, only attributes.

	You can initialize an Option with the following:

	func	The function that will be called when this
		option is processed on the command line.
		Calling sequence is:

			func(opt, arg)

		If there is no func, then this Option probably
		stores an optstring to be printed.

	helpline
		The string to be printed in -h output.  If no
		helpline is specified but a help string is
		specified (the usual case), a helpline will be
		constructed automatically from the short, long,
		arg, and help attributes.  (In practice, then,
		setting helpline without setting func allows you
		to print arbitrary lines of text in the -h
		output.)

	short	The string for short, single-hyphen
		command-line options.
		Do not include the hyphen:

			'a' for -a, 'xy' for -x and -y, etc.

	long	An array of strings for long, double-hyphen
		command-line options.  Do not include
		the hyphens:

			['my-option', 'verbose']

	arg	If this option takes an argument, this string
		specifies how you want it to appear in the
		-h output ('DIRECTORY', 'FILE', etc.).

	help	The help string that will be printed for
		this option in the -h output.  Must be
		49 characters or fewer.

	future	If non-zero, this indicates that this feature
		will be supported in a future release, not
		the currently planned one.  SCons will
		recognize the option, but it won't show up
		in the -h output.

	The following attribute is derived from the supplied attributes:

	optstring
		A string, with hyphens, describing the flags
		for this option, as constructed from the
		specified short, long and arg attributes.

	All Option objects are stored in the global option_list list,
	in the order in which they're created.  This is the list
	that's used to generate -h output, so the order in which the
	objects are created is the order in which they're printed.

	The upshot is that specifying a command-line option and having
	everything work correctly is a matter of defining a function to
	process its command-line argument (set the right flag, update
	the right value), and then creating an appropriate Option object
	at the correct point in the code below.
	"""

	def __init__(self, func = None, helpline = None,
		 short = None, long = None, arg = None,
		 help = None, future = None):
	    self.func = func
	    self.short = short
	    self.long = long
	    self.arg = arg
	    self.help = help
	    opts = []
	    if self.short:
		for c in self.short:
		    if arg:
			c = c + " " + arg
		    opts = opts + ['-' + c]
	    if self.long:
		l = self.long
		if arg:
		    l = map(lambda x,a=arg: x + "=" + a, self.long)
		opts = opts + map(lambda x: '--' + x, l)
	    self.optstring = string.join(opts, ', ')
	    if helpline:
		self.helpline = helpline
	    elif help and not future:
		if len(self.optstring) <= 26:
		    sep = " " * (28 - len(self.optstring))
		else:
		    sep = self.helpstring = "\n" + " " * 30
		self.helpline = "  " + self.optstring + sep + self.help
	    else:
		self.helpline = None
	    global option_list
	    option_list.append(self)

    # Generic routine for to-be-written options, used by multiple
    # options below.

    def opt_not_yet(opt, arg):
        sys.stderr.write("Warning:  the %s option is not yet implemented\n"
			  % opt)

    # In the following instantiations, the help string should be no
    # longer than 49 characters.  Use the following as a guide:
    #	help = "1234567890123456789012345678901234567890123456789"

    def opt_ignore(opt, arg):
	sys.stderr.write("Warning:  ignoring %s option\n" % opt)

    Option(func = opt_ignore,
	short = 'bmSt', long = ['no-keep-going', 'stop', 'touch'],
	help = "Ignored for compatibility.")

    def opt_c(opt, arg):
        global task_class, calc
        task_class = CleanTask
        class CleanCalculator:
            def bsig(self, node):
                return None
            def csig(self, node):
                return None
            def current(self, node, sig):
                return 0
            def write(self):
                pass
        calc = CleanCalculator()

    Option(func = opt_c,
	short = 'c', long = ['clean', 'remove'],
	help = "Remove specified targets and dependencies.")

    Option(func = opt_not_yet, future = 1,
	long = ['cache-disable', 'no-cache'],
	help = "Do not retrieve built targets from Cache.")

    Option(func = opt_not_yet, future = 1,
	long = ['cache-force', 'cache-populate'],
	help = "Copy already-built targets into the Cache.")

    Option(func = opt_not_yet, future = 1,
	long = ['cache-show'],
	help = "Print what would have built Cached targets.")

    def opt_C(opt, arg):
	try:
	    os.chdir(arg)
	except:
	    sys.stderr.write("Could not change directory to 'arg'\n")

    Option(func = opt_C,
	short = 'C', long = ['directory'], arg = 'DIRECTORY',
	help = "Change to DIRECTORY before doing anything.")

    Option(func = opt_not_yet,
	short = 'd',
	help = "Print file dependency information.")

    def opt_debug(opt, arg):
        global print_tree
        if arg == "pdb":
            args = [ sys.executable, "pdb.py" ] + \
                     filter(lambda x: x != "--debug=pdb", sys.argv)
            if sys.platform == 'win32':
                args[1] = os.path.join(sys.exec_prefix, "lib", "pdb.py")
                sys.exit(os.spawnve(os.P_WAIT, args[0], args, os.environ))
            else:
                args[1] = os.path.join(sys.exec_prefix,
                                       "lib",
                                       "python" + sys.version[0:3],
				       "pdb.py")
                os.execvpe(args[0], args, os.environ)
        elif arg == "tree":
            print_tree = 1
        else:
            sys.stderr.write("Warning:  %s is not a valid debug type\n"
                             % arg)

    Option(func = opt_debug,
           long = ['debug'], arg='TYPE',
           help = "Print various types of debugging information.")

    Option(func = opt_not_yet, future = 1,
	short = 'e', long = ['environment-overrides'],
	help = "Environment variables override makefiles.")

    def opt_f(opt, arg):
	global scripts
        scripts.append(arg)

    Option(func = opt_f,
	short = 'f', long = ['file', 'makefile', 'sconstruct'], arg = 'FILE',
	help = "Read FILE as the top-level SConstruct file.")

    def opt_help(opt, arg):
	global help_option
        help_option = 'h'
	SCons.Script.SConscript.print_help = 1

    Option(func = opt_help,
	short = 'h', long = ['help'],
	help = "Print defined help message, or this one.")

    def opt_help_options(opt, arg):
	global help_option
        help_option = 'H'

    Option(func = opt_help_options,
	short = 'H', long = ['help-options'],
	help = "Print this message and exit.")

    def opt_i(opt, arg):
        global ignore_errors
        ignore_errors = 1

    Option(func = opt_i,
	short = 'i', long = ['ignore-errors'],
	help = "Ignore errors from build actions.")

    def opt_I(opt, arg):
	global include_dirs
	include_dirs = include_dirs + [arg]

    Option(func = opt_I,
	short = 'I', long = ['include-dir'], arg = 'DIRECTORY',
	help = "Search DIRECTORY for imported Python modules.")

    def opt_j(opt, arg):
	global num_jobs
	try:
            num_jobs = int(arg)
	except:
            print UsageString()
            sys.exit(1)

	if num_jobs <= 0:
            print UsageString()
            sys.exit(1)

    Option(func = opt_j,
	short = 'j', long = ['jobs'], arg = 'N',
	help = "Allow N jobs at once.")

    def opt_k(opt, arg):
        global keep_going_on_error
        keep_going_on_error = 1

    Option(func = opt_k,
	short = 'k', long = ['keep-going'],
	help = "Keep going when a target can't be made.")

    Option(func = opt_not_yet, future = 1,
	short = 'l', long = ['load-average', 'max-load'], arg = 'N',
	help = "Don't start multiple jobs unless load is below N.")

    Option(func = opt_not_yet, future = 1,
	long = ['list-derived'],
	help = "Don't build; list files that would be built.")

    Option(func = opt_not_yet, future = 1,
	long = ['list-actions'],
	help = "Don't build; list files and build actions.")

    Option(func = opt_not_yet, future = 1,
	long = ['list-where'],
	help = "Don't build; list files and where defined.")

    def opt_n(opt, arg):
	SCons.Action.execute_actions = None

    Option(func = opt_n,
	short = 'n', long = ['no-exec', 'just-print', 'dry-run', 'recon'],
	help = "Don't build; just print commands.")

    Option(func = opt_not_yet, future = 1,
	short = 'o', long = ['old-file', 'assume-old'], arg = 'FILE',
	help = "Consider FILE to be old; don't rebuild it.")

    Option(func = opt_not_yet, future = 1,
	long = ['override'], arg = 'FILE',
	help = "Override variables as specified in FILE.")

    Option(func = opt_not_yet, future = 1,
	short = 'p',
	help = "Print internal environments/objects.")

    def opt_profile(opt, arg):
        sys.argv = filter(lambda x: x[0:10] != "--profile=", sys.argv)
        import profile
        profile.run('SCons.Script.main()', arg)

    Option(func = opt_profile,
	long = ['profile'], arg = 'FILE',
	help = "Profile SCons and put results in FILE.")

    def opt_q(opt, arg):
        global task_class
        task_class = QuestionTask

    Option(func = opt_q, future = 1,
	short = 'q', long = ['question'],
	help = "Don't build; exit status says if up to date.")

    Option(func = opt_not_yet, future = 1,
	short = 'rR', long = ['no-builtin-rules', 'no-builtin-variables'],
	help = "Clear default environments and variables.")

    Option(func = opt_not_yet, future = 1,
	long = ['random'],
	help = "Build dependencies in random order.")

    def opt_s(opt, arg):
	SCons.Action.print_actions = None

    Option(func = opt_s,
	short = 's', long = ['silent', 'quiet'],
	help = "Don't print commands.")

    def opt_u(opt, arg):
        global climb_up
        climb_up = 1

    Option(func = opt_u,
	short = 'u', long = ['up', 'search-up'],
	help = "Search up directory tree for SConstruct.")

    def opt_U(opt, arg):
        global climb_up
        climb_up = 2

    Option(func = opt_U,
	short = 'U',
	help = "Search up directory tree for SConstruct.")

    def option_v(opt, arg):
        import SCons
	print "SCons by Steven Knight et al.:"
	print "\tscript version 0.06"
	print "\tbuild engine version %s" % SCons.__version__
	print "Copyright 2001, 2002 Steven Knight"
	sys.exit(0)

    Option(func = option_v,
	short = 'v', long = ['version'],
	help = "Print the SCons version number and exit.")

    Option(func = opt_not_yet, future = 1,
	short = 'w', long = ['print-directory'],
	help = "Print the current directory.")

    Option(func = opt_not_yet, future = 1,
	long = ['no-print-directory'],
	help = "Turn off -w, even if it was turned on implicitly.")

    Option(func = opt_not_yet, future = 1,
	long = ['write-filenames'], arg = 'FILE',
	help = "Write all filenames examined into FILE.")

    Option(func = opt_not_yet, future = 1,
	short = 'W', long = ['what-if', 'new-file', 'assume-new'], arg = 'FILE',
	help = "Consider FILE to be changed.")

    Option(func = opt_not_yet, future = 1,
	long = ['warn-undefined-variables'],
	help = "Warn when an undefined variable is referenced.")

    Option(func = opt_not_yet, future = 1,
	short = 'Y', long = ['repository'], arg = 'REPOSITORY',
	help = "Search REPOSITORY for source and target files.")

    global short_opts
    global long_opts
    global opt_func
    for o in option_list:
	if o.short:
	    if o.func:
		for c in o.short:
		    opt_func['-' + c] = o.func
	    short_opts = short_opts + o.short
	    if o.arg:
		short_opts = short_opts + ":"
	if o.long:
	    if o.func:
		for l in o.long:
		    opt_func['--' + l] = o.func
	    if o.arg:
		long_opts = long_opts + map(lambda a: a + "=", o.long)
	    else:
		long_opts = long_opts + o.long

options_init()



def _SConstruct_exists(dirname=''):
    """This function checks that an SConstruct file exists in a directory.
    If so, it returns the path of the file. By default, it checks the
    current directory.
    """
    for file in ['SConstruct', 'Sconstruct', 'sconstruct']:
        sfile = os.path.join(dirname, file)
        if os.path.isfile(sfile):
            return sfile
    return None


def UsageString():
    help_opts = filter(lambda x: x.helpline, option_list)
    s = "Usage: scons [OPTION] [TARGET] ...\n" + "Options:\n" + \
	string.join(map(lambda x: x.helpline, help_opts), "\n") + "\n"
    return s



def _main():
    global scripts, num_jobs, task_class, calc, target_top

    targets = []

    # It looks like 2.0 changed the name of the exception class
    # raised by getopt.
    try:
	getopt_err = getopt.GetoptError
    except:
	getopt_err = getopt.error

    try:
	cmd_opts, t = getopt.getopt(string.split(os.environ['SCONSFLAGS']),
					  short_opts, long_opts)
    except KeyError:
	# It's all right if there's no SCONSFLAGS environment variable.
	pass
    except getopt_err, x:
	_scons_user_warning("SCONSFLAGS " + str(x))
    else:
	for opt, arg in cmd_opts:
	    opt_func[opt](opt, arg)

    try:
        cmd_opts, args = getopt.getopt(sys.argv[1:], short_opts, long_opts)
    except getopt_err, x:
	_scons_user_error(x)
    else:
	for opt, arg in cmd_opts:
	    opt_func[opt](opt, arg)
        xmit_args = []
        for a in args:
            if '=' in a:
                xmit_args.append(a)
            else:
                targets.append(a)
        SCons.Script.SConscript._scons_add_args(xmit_args)

    if climb_up:
        target_top = ''  # directory to prepend to targets
        script_dir = os.getcwd()  # location of script
        while script_dir and not _SConstruct_exists(script_dir):
            script_dir, last_part = os.path.split(script_dir)
            if last_part:
                target_top = os.path.join(last_part, target_top)
            else:
                script_dir = ''
        if script_dir:
	    print "scons: Entering directory %s" % script_dir
            os.chdir(script_dir)
        else:
            raise UserError, "No SConstruct file found."
        
    SCons.Node.FS.default_fs.set_toplevel_dir(os.getcwd())

    if not scripts:
        sfile = _SConstruct_exists()
        if sfile:
            scripts.append(sfile)

    if help_option == 'H':
	print UsageString()
	sys.exit(0)

    if not scripts:
        if help_option == 'h':
            # There's no SConstruct, but they specified -h.
            # Give them the options usage now, before we fail
            # trying to read a non-existent SConstruct file.
	    print UsageString()
	    sys.exit(0)
	else:
	    raise UserError, "No SConstruct file found."

    class Unbuffered:
        def __init__(self, file):
            self.file = file
        def write(self, arg):
            self.file.write(arg)
            self.file.flush()
        def __getattr__(self, attr):
            return getattr(self.file, attr)

    sys.stdout = Unbuffered(sys.stdout)

    sys.path = include_dirs + sys.path

    for script in scripts:
        SCons.Script.SConscript.SConscript(script)

    SCons.Node.FS.default_fs.chdir(SCons.Node.FS.default_fs.Top)

    if help_option == 'h':
	# They specified -h, but there was no Help() inside the
	# SConscript files.  Give them the options usage.
	print UsageString()
	sys.exit(0)

    if target_top:
        if climb_up == 2 and not targets:
            # -U with default targets
            target_top = None
        else:
            target_top = SCons.Node.FS.default_fs.Dir(target_top)

    if not targets:
        targets = SCons.Script.SConscript.default_targets

    def Entry(x, top = target_top):
        if isinstance(x, SCons.Node.Node):
            node = x
        else:
            try:
                node = SCons.Node.Alias.default_ans.lookup(x)
                if node is None:
                    node = SCons.Node.FS.default_fs.Entry(x,
                                                          directory = top,
                                                          create = 0)
            except UserError:
                string = "scons: *** Do not know how to make target `%s'." % x
                if not keep_going_on_error:
                    sys.stderr.write(string + "  Stop.\n")
                    sys.exit(2)
                sys.stderr.write(string + "\n")
                node = None
        if top and not node.is_under(top):
            if isinstance(node, SCons.Node.FS.Dir) and top.is_under(node):
                node = top
            else:
                node = None
        return node

    nodes = filter(lambda x: x is not None, map(Entry, targets))

    if not calc:
        calc = SCons.Sig.Calculator(SCons.Sig.MD5)

    taskmaster = SCons.Taskmaster.Taskmaster(nodes, task_class, calc)

    jobs = SCons.Job.Jobs(num_jobs, taskmaster)
    jobs.start()
    jobs.wait()

    SCons.Sig.write()

def main():
    try:
	_main()
    except SystemExit:
        pass
    except KeyboardInterrupt:
        print "Build interrupted."
        sys.exit(1)
    except SyntaxError, e:
        _scons_syntax_error(e)
    except UserError, e:
        _scons_user_error(e)
    except:
        _scons_other_errors()

    sys.exit(exit_status)
